import torch
import operator
import torchdata.datapipes.iter as it

from .data import FileFormat, build_input_data_pipe


from .models import Kitsune
from .data import FileFormat, build_input_data_pipe
from .engine import build_feature_mapper, train_single_epoch, predict

class KitModel:
    def __init__(self):
        self.fm = None
        self.model = None
        self.optimizer = None
        self.dp = None

    # Generate Data
    def __gen_dp__(self, data_path):
        return build_input_data_pipe(
            data_path, batch_size=32, shuffle=True, file_format=FileFormat.csv)

    # feature mapper
    def __map_features__(self):
        return build_feature_mapper(self.dp, 100)

    # load and store model
    def load_model(self, model_path:str):
        self.model = Kitsune.from_pretrained(model_path)

    def store_model(self, model_path:StopAsyncIteration()):
        self.model.save(model_path)

    def score(self, x: torch.Tensor):
        return self.model.score(x)

    # train model
    def train_model(self, input_path: str,
                    batch_size: int = 32,
                    file_format: FileFormat = "csv",
                    compression_rate: float = 0.6):

        # Build data pipe
        self.dp = self.__gen_dp__(input_path)

        # Build feature mapper
        self.fm = self.__map_features__()

        # Build model
        self.model = Kitsune(feature_mapper=self.fm,
                        compression_rate=compression_rate)
        
        # Build optimizer
        self.optimizer = torch.optim.SGD(self.model.parameters(), lr=1e-3)
        
        
        # TODO: take epoch as input
        for epoch in range(10):
            train_single_epoch(
                self.model, self.dp, self.optimizer, epoch=epoch)
        

    # test model
    def test(self, input_path: str, batch_size: int = 32, file_format: FileFormat = "csv"):
        test_data = self.__gen_dp__(input_path)   
        scores = predict(self.model, test_data)

        return scores



import numpy as np
# correct the below import using the correct relative path.

from . import dA as AE
from . import corClust as CC

# import dA as AE
# import corClust as CC

# This class represents a KitNET machine learner.
# KitNET is a lightweight online anomaly detection algorithm based on an ensemble of autoencoders.
# For more information and citation, please see our NDSS'18 paper: Kitsune: An Ensemble of Autoencoders for Online Network Intrusion Detection
# For licensing information, see the end of this document


class KitNET:
    # n: the number of features in your input dataset (i.e., x \in R^n)
    # m: the maximum size of any autoencoder in the ensemble layer
    # AD_grace_period: the number of instances the network will learn from before producing anomaly scores
    # FM_grace_period: the number of instances which will be taken to learn the feature mapping. If 'None', then FM_grace_period=AM_grace_period
    # learning_rate: the default stochastic gradient descent learning rate for all autoencoders in the KitNET instance.
    # hidden_ratio: the default ratio of hidden to visible neurons. E.g., 0.75 will cause roughly a 25% compression in the hidden layer.
    # feature_map: One may optionally provide a feature map instead of learning one. The map must be a list,
    #           where the i-th entry contains a list of the feature indices to be assingned to the i-th autoencoder in the ensemble.
    #           For example, [[2,5,3],[4,0,1],[6,7]]
    def __init__(self, n, max_autoencoder_size=10, FM_grace_period=None, AD_grace_period=10000, learning_rate=0.1, hidden_ratio=0.75, feature_map=None, normalize=True, input_precision=None, quantize=None):
        # Parameters:
        self.AD_grace_period = AD_grace_period
        if FM_grace_period is None:
            self.FM_grace_period = AD_grace_period
        else:
            self.FM_grace_period = FM_grace_period
        self.input_precision = input_precision
        if max_autoencoder_size <= 0:
            self.m = 1
        else:
            self.m = max_autoencoder_size
        self.lr = learning_rate
        self.hr = hidden_ratio
        self.n = n
        self.normalize = normalize
        # Variables
        self.n_trained = 0  # the number of training instances so far
        self.n_executed = 0  # the number of executed instances so far
        self.v = feature_map
        self.ensembleLayer = []
        self.outputLayer = None
        self.quantize = quantize
        if self.v is None:
            pass
            print("Feature-Mapper: train-mode, Anomaly-Detector: off-mode")
        else:
            self.__createAD__()
            print("Feature-Mapper: execute-mode, Anomaly-Detector: train-mode")
        # incremental feature cluatering for the feature mapping process
        self.FM = CC.corClust(self.n)

    # If FM_grace_period+AM_grace_period has passed, then this function executes KitNET on x. Otherwise, this function learns from x.
    # x: a numpy array of length n
    # Note: KitNET automatically performs 0-1 normalization on all attributes.
    def process(self, x):
        # if all -1 it means the packet was ignored
        if x.all() == -1:
            return 0.

        if self.n_trained > self.FM_grace_period + self.AD_grace_period:  # If both the FM and AD are in execute-mode
            return self.execute(x)
        else:
            self.train(x)
            return 0.0

    # alias so it is compatible with sklearn models, input and output are all 2d arrays
    def decision_function(self, x):
        anom_score = self.process(x[0])
        return [-anom_score]

    # alias for execute for it is compatible with tf models, processes in batches
    def predict(self, x):
        return np.array([self.process(x[i]) for i in range(np.array(x).shape[0])])

    # force train KitNET on x
    # returns the anomaly score of x during training (do not use for alerting)
    def train(self, x):
        # If the FM is in train-mode, and the user has not supplied a feature mapping
        if self.n_trained <= self.FM_grace_period and self.v is None:
            # update the incremetnal correlation matrix
            self.FM.update(x)
            if self.n_trained == self.FM_grace_period:  # If the feature mapping should be instantiated
                self.v = self.FM.cluster(self.m)
                self.__createAD__()
                print("The Feature-Mapper found a mapping: "+str(self.n)+" features to "+str(len(self.v))+" autoencoders.")
                print("Feature-Mapper: execute-mode, Anomaly-Detector: train-mode")
        else:  # train
            # Ensemble Layer
            S_l1 = np.zeros(len(self.ensembleLayer))
            for a in range(len(self.ensembleLayer)):
                # make sub instance for autoencoder 'a'
                xi = x[self.v[a]]
                S_l1[a] = self.ensembleLayer[a].train(xi)
            # OutputLayer
            self.outputLayer.train(S_l1)
            if self.n_trained == self.AD_grace_period + self.FM_grace_period:
                pass
                print("Feature-Mapper: execute-mode, Anomaly-Detector: execute-mode")
        self.n_trained += 1

    # force execute KitNET on x
    def execute(self, x):
        if self.v is None:
            raise RuntimeError(
                'KitNET Cannot execute x, because a feature mapping has not yet been learned or provided. Try running process(x) instead.')
        else:
            self.n_executed += 1
            # Ensemble Layer
            S_l1 = np.zeros(len(self.ensembleLayer))
            for a in range(len(self.ensembleLayer)):
                # make sub inst
                xi = x[self.v[a]]
                S_l1[a] = self.ensembleLayer[a].execute(xi)
            # OutputLayer
            return self.outputLayer.execute(S_l1)

    def __createAD__(self):
        # construct ensemble layer
        for map in self.v:
            params = AE.dA_params(n_visible=len(map), n_hidden=0, lr=self.lr, corruption_level=0, gracePeriod=0,
                                  hiddenRatio=self.hr, normalize=self.normalize, input_precision=self.input_precision, quantize=self.quantize)
            self.ensembleLayer.append(AE.dA(params))

        # construct output layer
        params = AE.dA_params(len(self.v), n_hidden=0, lr=self.lr, corruption_level=0, gracePeriod=0, hiddenRatio=self.hr,
                              normalize=self.normalize, quantize=self.quantize, input_precision=self.input_precision)
        self.outputLayer = AE.dA(params)

    def get_params(self):
        return_dict = {"ensemble": []}
        for i in range(len(self.ensembleLayer)):
            return_dict["ensemble"].append(self.ensembleLayer[i].get_params())
        return_dict["output"] = self.outputLayer.get_params()
        return return_dict

    def set_params(self, new_param):
        for i in range(len(new_param["ensemble"])):
            self.ensembleLayer[i].set_params(new_param["ensemble"][i])
        self.outputLayer.set_params(new_param["output"])


# Copyright (c) 2017 Yisroel Mirsky
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.



import math
import numpy as np

# MIT License
#
# Copyright (c) 2018 Yisroel mirsky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


#compile with:   python setup.py build_ext --inplace

import pyximport; pyximport.install()

cdef class incStat:
    cdef str ID
    cdef double CF1
    cdef double CF2
    cdef double w
    cdef int isTypeDiff
    cdef double Lambda
    cdef double lastTimestamp
    cdef double cur_mean
    cdef double cur_var
    cdef double cur_std
    cdef list covs

    def __init__(self, double Lambda, str ID, double init_time=0, int isTypeDiff=False):  # timestamp is creation time
        self.ID = ID
        self.CF1 = 0  # linear sum
        self.CF2 = 0  # sum of squares
        self.w = 1e-20  # weight
        self.isTypeDiff = isTypeDiff
        self.Lambda = Lambda  # Decay Factor
        self.lastTimestamp = init_time
        self.cur_mean = np.nan
        self.cur_var = np.nan
        self.cur_std = np.nan
        self.covs = [] # a list of incStat_covs (references) with relate to this incStat

    cdef insert(self, double v, double t=0):  # v is a scalar, t is v's arrival the timestamp
        if self.isTypeDiff:
            if t - self.lastTimestamp > 0:
                v = t - self.lastTimestamp
            else:
                v = 0
        self.processDecay(t)

        # update with v
        self.CF1 += v
        self.CF2 += math.pow(v, 2)
        self.w += 1
        self.cur_mean = np.nan  # force recalculation if called
        self.cur_var = np.nan
        self.cur_std = np.nan

        # update covs (if any)
        cdef incStat_cov cov
        for c in self.covs:
            cov = c
            cov.update_cov(self.ID, v, t)

    cdef processDecay(self, double timestamp):
        cdef double factor, timeDiff
        factor = 1
        # check for decay
        timeDiff = timestamp - self.lastTimestamp
        if timeDiff > 0:
            factor = math.pow(2, (-self.Lambda * timeDiff))
            self.CF1 = self.CF1 * factor
            self.CF2 = self.CF2 * factor
            self.w = self.w * factor
            self.lastTimestamp = timestamp
        return factor

    cdef weight(self):
        return self.w

    cdef mean(self):
        if math.isnan(self.cur_mean):  # calculate it only once when necessary
            self.cur_mean = self.CF1 / self.w
        return self.cur_mean

    cdef var(self):
        if math.isnan(self.cur_var):  # calculate it only once when necessary
            self.cur_var = abs(self.CF2 / self.w - math.pow(self.mean(), 2))
        return self.cur_var

    cdef std(self):
        if math.isnan(self.cur_std):  # calculate it only once when necessary
            self.cur_std = math.sqrt(self.var())
        return self.cur_std

    cdef cov(self,ID2):
        for cov in self.covs:
            if cov.isRelated(ID2):
                return cov.cov()
        return [np.nan]

    cdef pcc(self,ID2):
        for cov in self.covs:
            if cov.isRelated(ID2):
                return cov.pcc()
        return [np.nan]

    cdef cov_pcc(self,ID2):
        cdef incStat_cov cov
        for c in self.covs:
            cov = c
            if cov.isRelated(ID2):
                return cov.get_stats1()
        return [np.nan]*2

    cdef radius(self, other_incStats):  # the radius of a set of incStats
        cdef double A
        A = self.var()
        cdef incStat incSc
        for incS in other_incStats:
            incSc = incS
            A += incSc.var()
        return math.sqrt(A)

    cdef magnitude(self, other_incStats):  # the magnitude of a set of incStats
        cdef double A
        A = math.pow(self.mean(), 2)
        cdef incStat incSc
        for incS in other_incStats:
            incSc = incS
            A += math.pow(incSc.mean(), 2)
        return math.sqrt(A)

    #calculates and pulls all stats on this stream
    cdef allstats_1D(self):
        self.cur_mean = self.CF1 / self.w
        self.cur_var = abs(self.CF2 / self.w - math.pow(self.cur_mean, 2))
        return [self.w, self.cur_mean, self.cur_var]

    #calculates and pulls all stats on this stream, and stats shared with the indicated stream
    cdef allstats_2D(self, str ID2):
        stats1D = self.allstats_1D()
        # Find cov component
        stats2D = [np.nan] * 4
        cdef incStat_cov cov
        for c in self.covs:
            cov = c
            if cov.isRelated(ID2):
                stats2D = cov.get_stats2()
                break
        return stats1D + stats2D

    cdef getHeaders_1D(self, suffix=True):
        if self.ID is None:
            s0=""
        else:
            s0 = "_0"
        if suffix:
            s0 = "_"+self.ID
        headers = ["weight"+s0, "mean"+s0, "std"+s0]
        return headers

    cdef getHeaders_2D(self, ID2, suffix=True):
        hdrs1D = self.getHeaders_1D(suffix)
        if self.ID is None:
            s0=""
            s1=""
        else:
            s0 = "_0"
            s1 = "_1"
        if suffix:
            s0 = "_"+self.ID
            s1 = "_" + ID2
        hdrs2D = ["radius_" + s0 + "_" + s1, "magnitude_" + s0 + "_" + s1, "covariance_" + s0 + "_" + s1,
                   "pcc_" + s0 + "_" + s1]
        return hdrs1D+hdrs2D

    # def toJSON(self):
    #     j = {}
    #     j['CF1'] = self.CF1
    #     j['CF2'] = self.CF2
    #     j['w'] = self.w
    #     j['isTypeDiff'] = self.isTypeDiff
    #     j['Lambda'] = self.Lambda
    #     j['lastTimestamp'] = self.lastTimestamp
    #     return json.dumps(j)
    #
    # def loadFromJSON(self,JSONstring):
    #     j = json.loads(JSONstring)
    #     self.CF1 = j['CF1']
    #     self.CF2 = j['CF2']
    #     self.w = j['w']
    #     self.isTypeDiff = j['isTypeDiff']
    #     self.Lambda = j['Lambda']
    #     self.lastTimestamp = j['lastTimestamp']

#like incStat, but maintains stats between two streams
#TODO: make it possble to call incstat magnitude and raduis withour list of incstsats (just single incstat objects) for cov.getstats2 typcast call
cdef class incStat_cov:
    cdef double CF3
    cdef double w3
    cdef double lastTimestamp_cf3
    cdef incStat incS1
    cdef incStat incS2
    cdef extrapolator ex1
    cdef extrapolator ex2


    def __init__(self, incStat incS1,incStat incS2, double init_time = 0):
        # store references tot he streams' incStats
        self.incS1 = incS1
        self.incS2 = incS2

        # init extrapolators
        self.ex1 = extrapolator()
        self.ex2 = extrapolator()

        # init sum product residuals
        self.CF3 = 0 # sum of residule products (A-uA)(B-uB)
        self.w3 = 1e-20
        self.lastTimestamp_cf3 = init_time

    #other_incS_decay is the decay factor of the other incstat
    # ID: the stream ID which produced (v,t)
    cdef update_cov(self, str ID, double v, double t):  # it is assumes that incStat "ID" has ALREADY been updated with (t,v) [this si performed automatically in method incStat.insert()]
        # find incStat
        cdef int inc
        if ID == self.incS1.ID:
            inc = 0
        else:
            inc = 1

        # Decay residules
        self.processDecay(t)

        # Update extrapolator for current stream AND
        # Extrapolate other stream AND
        # Compute and update residule
        cdef double v_other
        if inc == 0:
            self.ex1.insert(t,v)
            v_other = self.ex2.predict(t)
            self.CF3 += (v - self.incS1.mean()) * (v_other - self.incS2.mean())
        else:
            self.ex2.insert(t,v)
            v_other = self.ex1.predict(t)
            self.CF3 += (v - self.incS2.mean()) * (v_other - self.incS1.mean())
        self.w3 += 1

    cdef processDecay(self,double t):
        cdef double factor
        factor = 1
        # check for decay cf3
        cdef double timeDiffs_cf3
        timeDiffs_cf3 = t - self.lastTimestamp_cf3
        if timeDiffs_cf3 > 0:
            factor = math.pow(2, (-(self.incS1.Lambda) * timeDiffs_cf3))
            self.CF3 *= factor
            self.w3 *= factor
            self.lastTimestamp_cf3 = t
        return factor

    #todo: add W3 for cf3

    #covariance approximation
    cdef cov(self):
        return self.CF3 / self.w3

    # Pearson corl. coef
    cdef pcc(self):
        cdef double ss
        ss = self.incS1.std() * self.incS2.std()
        if ss != 0:
            return self.cov() / ss
        else:
            return 0

    def isRelated(self, str ID):
        if self.incS1.ID == ID or self.incS2.ID == ID:
            return True
        else:
            return False

    # calculates and pulls all correlative stats
    cdef get_stats1(self):
        return [self.cov(), self.pcc()]

    # calculates and pulls all correlative stats AND 2D stats from both streams (incStat)
    cdef get_stats2(self):
        return [self.incS1.radius([self.incS2]),self.incS1.magnitude([self.incS2]),self.cov(), self.pcc()]

    # calculates and pulls all correlative stats AND 2D stats AND the regular stats from both streams (incStat)
    cdef get_stats3(self):
        return [self.incS1.w,self.incS1.mean(),self.incS1.std(),self.incS2.w,self.incS2.mean(),self.incS2.std(),self.cov(), self.pcc()]

    # calculates and pulls all correlative stats AND the regular stats from both incStats AND 2D stats
    cdef get_stats4(self):
        return [self.incS1.w,self.incS1.mean(),self.incS1.std(),self.incS2.w,self.incS2.mean(),self.incS2.std(), self.incS1.radius([self.incS2]),self.incS1.magnitude([self.incS2]),self.cov(), self.pcc()]

    cdef getHeaders(self,int ver,int suffix=True): #ver = {1,2,3,4}
        headers = []
        s0 = "0"
        s1 = "1"
        if suffix:
            s0 = self.incS1.ID
            s1 = self.incS2.ID

        if ver == 1:
            headers = ["covariance_"+s0+"_"+s1, "pcc_"+s0+"_"+s1]
        if ver == 2:
            headers = ["radius_"+s0+"_"+s1, "magnitude_"+s0+"_"+s1, "covariance_"+s0+"_"+s1, "pcc_"+s0+"_"+s1]
        if ver == 3:
            headers = ["weight_"+s0, "mean_"+s0, "std_"+s0,"weight_"+s1, "mean_"+s1, "std_"+s1, "covariance_"+s0+"_"+s1, "pcc_"+s0+"_"+s1]
        if ver == 4:
            headers = ["weight_" + s0, "mean_" + s0, "std_" + s0, "covariance_" + s0 + "_" + s1, "pcc_" + s0 + "_" + s1]
        if ver == 5:
            headers = ["weight_"+s0, "mean_"+s0, "std_"+s0,"weight_"+s1, "mean_"+s1, "std_"+s1, "radius_"+s0+"_"+s1, "magnitude_"+s0+"_"+s1, "covariance_"+s0+"_"+s1, "pcc_"+s0+"_"+s1]
        return headers


cdef class incStatDB:
    cdef double limit
    cdef double df_lambda
    cdef dict HT

    # default_lambda: use this as the lambda for all streams. If not specified, then you must supply a Lambda with every query.
    def __init__(self,double limit=np.Inf,double default_lambda=np.nan):
        self.HT = dict()
        self.limit = limit
        self.df_lambda = default_lambda

    cdef get_lambda(self,double Lambda):
        if not np.isnan(self.df_lambda):
            Lambda = self.df_lambda
        return Lambda

    # Registers a new stream. init_time: init lastTimestamp of the incStat
    def register(self,str ID,double Lambda=1,double init_time=0,int isTypeDiff=False):
        #Default Lambda?
        Lambda = self.get_lambda(Lambda)

        #Retrieve incStat
        cdef str key
        key = ID+"_"+str(Lambda)

        cdef incStat incS
        incS = self.HT.get(key)
        if incS is None: #does not already exist
            if len(self.HT) + 1 > self.limit:
                raise LookupError(
                    'Adding Entry:\n' + key + '\nwould exceed incStatHT 1D limit of ' + str(
                        self.limit) + '.\nObservation Rejected.')
            incS = incStat(Lambda, ID, init_time, isTypeDiff)
            self.HT[key] = incS #add new entry
        return incS

    # Registers covariance tracking for two streams, registers missing streams
    def register_cov(self,str ID1, str ID2, double Lambda=1, double init_time=0, int isTypeDiff=False):
        #Default Lambda?
        Lambda = self.get_lambda(Lambda)

        # Lookup both streams
        cdef incStat incS1
        cdef incStat incS2
        incS1 = self.register(ID1,Lambda,init_time,isTypeDiff)
        incS2 = self.register(ID2,Lambda,init_time,isTypeDiff)

        #check for pre-exiting link
        for cov in incS1.covs:
            if cov.isRelated(ID2):
                return cov #there is a pre-exiting link

        # Link incStats
        inc_cov = incStat_cov(incS1,incS2,init_time)
        incS1.covs.append(inc_cov)
        incS2.covs.append(inc_cov)
        return inc_cov

    # updates/registers stream
    def update(self,str ID,double t,double v,double Lambda=1,int isTypeDiff=False):
        cdef incStat incS
        incS = self.register(ID,Lambda,t,isTypeDiff)
        incS.insert(v,t)
        return incS

    # Pulls current stats from the given ID
    def get_1D_Stats(self,str ID,double Lambda=1): #weight, mean, std
        #Default Lambda?
        Lambda = self.get_lambda(Lambda)

        #Get incStat
        cdef incStat incS
        incS = self.HT.get(ID+"_"+str(Lambda))
        if incS is None:  # does not already exist
            return [np.na]*3
        else:
            return incS.allstats_1D()

    # Pulls current correlational stats from the given IDs
    def get_2D_Stats(self, str ID1, str ID2, double Lambda=1): #cov, pcc
        # Default Lambda?
        Lambda = self.get_lambda(Lambda)

        # Get incStat
        cdef incStat incS
        incS = self.HT.get(ID1 + "_" + str(Lambda))
        if incS is None:  # does not exist
            return [np.na]*2

        # find relevant cov entry
        return incS.cov_pcc(ID2)

    # Pulls all correlational stats registered with the given ID
    # returns tuple [0]: stats-covs&pccs, [2]: IDs
    def get_all_2D_Stats(self, str ID, double Lambda=1):  # cov, pcc
        # Default Lambda?
        Lambda = self.get_lambda(Lambda)

        # Get incStat
        cdef incStat incS1
        incS1 = self.HT.get(ID + "_" + str(Lambda))
        if incS1 is None:  # does not exist
            return ([],[])

        # find relevant cov entry
        stats = []
        IDs = []
        for cov in incS1.covs:
            stats.append(cov.get_stats1())
            IDs.append([cov.incS1.ID,cov.incS2.ID])
        return stats,IDs

    # Pulls current multidimensional stats from the given IDs
    def get_nD_Stats(self,IDs,double Lambda=1): #radius, magnitude (IDs is a list)
        # Default Lambda?
        Lambda = self.get_lambda(Lambda)

        # Get incStats
        incStats = []
        for ID in IDs:
            incS = self.HT.get(ID + "_" + str(Lambda))
            if incS is not None:  #exists
                incStats.append(incS)

        # Compute stats
        cdef double rad, mag
        rad = 0 #radius
        mag = 0 #magnitude
        for incS in incStats:
            rad += incS.var()
            mag += incS.mean()**2

        return [np.sqrt(rad),np.sqrt(mag)]

    # Updates and then pulls current 1D stats from the given ID. Automatically registers previously unknown stream IDs
    def update_get_1D_Stats(self, str ID,double t, double v, double Lambda=1, int isTypeDiff=False):  # weight, mean, std
        cdef incStat incS
        incS = self.update(ID,t,v,Lambda,isTypeDiff)
        return incS.allstats_1D()


    # Updates and then pulls current correlative stats between the given IDs. Automatically registers previously unknown stream IDs, and cov tracking
    #Note: AfterImage does not currently support Diff Type streams for correlational statistics.
    def update_get_2D_Stats(self, str ID1, str ID2,double t1,double v1,double Lambda=1, int level=1):  #level=  1:cov,pcc  2:radius,magnitude,cov,pcc
        #retrieve/add cov tracker
        cdef incStat_cov inc_cov
        inc_cov = self.register_cov(ID1, ID2, Lambda,  t1)
        # Update cov tracker
        inc_cov.update_cov(ID1,v1,t1)
        if level == 1:
            return inc_cov.get_stats1()
        else:
            return inc_cov.get_stats2()

    # Updates and then pulls current 1D and 2D stats from the given IDs. Automatically registers previously unknown stream IDs
    def update_get_1D2D_Stats(self, str ID1, str ID2, double t1,double v1,double Lambda=1):  # weight, mean, std
        return self.update_get_1D_Stats(ID1,t1,v1,Lambda) + self.update_get_2D_Stats(ID1,ID2,t1,v1,Lambda,level=2)

    def getHeaders_1D(self,Lambda=1,ID=''):
        # Default Lambda?
        cdef double L
        L = Lambda
        L = self.get_lambda(L)
        hdrs = incStat(L,ID).getHeaders_1D(suffix=False)
        return [str(L)+"_"+s for s in hdrs]

    def getHeaders_2D(self,Lambda=1,IDs=None, ver=1): #IDs is a 2-element list or tuple
        # Default Lambda?
        cdef double L
        L = Lambda
        L = self.get_lambda(L)
        if IDs is None:
            IDs = ['0','1']
        hdrs = incStat_cov(incStat(L,IDs[0]),incStat(L,IDs[0]),L).getHeaders(ver,suffix=False)
        return [str(Lambda)+"_"+s for s in hdrs]

    def getHeaders_1D2D(self,Lambda=1,IDs=None, ver=1):
        # Default Lambda?
        cdef double L
        L = Lambda
        L = self.get_lambda(L)
        if IDs is None:
            IDs = ['0','1']
        hdrs1D = self.getHeaders_1D(L,IDs[0])
        hdrs2D = self.getHeaders_2D(L,IDs, ver)
        return hdrs1D + hdrs2D

    def getHeaders_nD(self,Lambda=1,IDs=[]): #IDs is a n-element list or tuple
        # Default Lambda?
        cdef double L
        L = Lambda
        ID = ":"
        for s in IDs:
            ID += "_"+s
        L = self.get_lambda(L)
        hdrs = ["radius"+ID, "magnitude"+ID]
        return [str(L)+"_"+s for s in hdrs]


    #cleans out records that have a weight less than the cutoff.
    #returns number or removed records.
    def cleanOutOldRecords(self,double cutoffWeight,double curTime):
        cdef int n
        cdef double W
        n = 0
        dump = sorted(self.HT.items(), key=lambda tup: tup[1][0].getMaxW(curTime))
        for entry in dump:
            entry[1][0].processDecay(curTime)
            W = entry[1][0].w
            if W <= cutoffWeight:
                key = entry[0]
                del entry[1][0]
                del self.HT[key]
                n=n+1
            elif W > cutoffWeight:
                break
        return n





class incHist:
    #ubIsAnom means that the HBOS score for vals that fall past the upped bound are Inf (not 0)
    def __init__(self,nbins,Lambda=0,ubIsAnom=True,lbIsAnom=True,lbound=-10,ubound=10,scaleGrace=None):
        self.scaleGrace = scaleGrace #the numbe rof instances to observe until a range it determeined
        if scaleGrace is not None:
            self.lbound = np.Inf
            self.ubound = -np.Inf
            self.binSize = None
            self.isScaling = True
        else:
            self.lbound = lbound
            self.ubound = ubound
            self.binSize = (ubound - lbound)/nbins
            self.isScaling = False
        self.nbins = nbins
        self.ubIsAnom = ubIsAnom
        self.lbIsAnom = lbIsAnom
        self.n = 0

        self.Lambda = Lambda
        self.W = np.zeros(nbins)
        self.lT = np.zeros(nbins) #last timestamp of each respective bin
        self.tallestBin = 0 #indx to the bin that currently has the largest freq weight (assumed...)

    #assumes even bin width starting from lbound until ubound. beyond bounds are assigned to the closest bin
    def getBinIndx(self,val,win=0):
        indx = int(np.floor((val - self.lbound)/self.binSize))
        if win == 0:
            if indx < 0:
                return -np.Inf
            if indx > (self.nbins - 1):
                return np.Inf
            return indx
        else: #windowed Histogram
            if indx - win < 0: #does the left of the window stick out of bounds?
                if indx + win >= 0: #if yes, then is there some overlap with inbounds?
                    return range(0,indx+win+1) #return the inbounds range
                else: #then the entire window is our of bounds to the left
                    return -np.Inf
            if indx + win > self.nbins - 1: #does the right of the window stick out of bounds?
                if indx - win < self.nbins: #if yes, then is there some overlap with inbounds?
                    return range(indx - win,self.nbins) #return the inbounds range
                else: #then the entire window is our of bounds to the right
                    return np.Inf
            return range(indx-win,indx+win+1)


    def processDecay(self, bin, timestamp):
        # check for decay
        timeDiff = timestamp - self.lT[bin]
        if np.isscalar(timeDiff):
            if timeDiff > 0:
                factor = math.pow(2, (-self.Lambda * timeDiff))
                self.W[bin] = self.W[bin] * factor
                self.lT[bin] = timestamp
        else: #array
            timeDiff[timeDiff<0]=0 #don't affect decay of out of order entries
            factor = np.power(2, (-self.Lambda * timeDiff))
            #b4 = self.W[bin]
            self.W[bin] = self.W[bin] * factor
            self.lT[bin] = timestamp

    def insert(self,val,timestamp,penalty=False):
        self.n = self.n + 1
        if self.isScaling:
            if self.n < self.scaleGrace:
                if self.lbound > val:
                    self.lbound = val
                if self.ubound < val:
                    self.ubound = val
            if self.n == self.scaleGrace:
                if self.ubound == self.lbound:
                    self.scaleGrace = self.scaleGrace + 1000
                else:
                    width = self.ubound - self.lbound
                    self.ubound = self.ubound + width
                    self.lbound = self.lbound - width
                    self.binSize = (self.ubound - self.lbound) / self.nbins
                    self.isScaling = False
        else:
            bin = self.getBinIndx(val)
            if not np.isinf(bin): #
                self.processDecay(bin, timestamp)
                if penalty:
                    tallestW = self.W[self.tallestBin]
                    scale = tallestW if tallestW > 0 else 1
                    fn = self.W[bin]/scale
                    inc = self.halfsigmoid(fn+0.005,-1.03)
                else:
                    inc = 1
                self.W[bin] = self.W[bin] + inc
                #track who has the tallest bin (for normilization)
                if self.W[bin] > self.W[self.tallestBin]:
                    self.tallestBin = bin

    def halfsigmoid(self,x,k):
        return (k*x)/(k-x+1)

    def score(self,val,timestamp=-1,win=0): #HBOS for one dimension
        if self.isScaling:
            return 0.0
        else:
            bin = self.getBinIndx(val,win=win)
            if np.isscalar(bin):
                if np.isinf(bin):
                    if self.ubIsAnom and bin > 0:
                        return np.Inf #it's an anomaly because it passes the upper bound
                    elif self.lbIsAnom and bin < 0:
                        return np.Inf  # it's an anomaly because it passes the lower bound
                    else:
                        return 0.0 #it fell outside a bound which is consedered not anomalous
            self.processDecay(bin,timestamp) #if timestamp = -1, no decay will be applied
            w = np.mean(self.W[bin])
            if w == 0:
                return np.Inf  # no stat history, anomaly!
            else:
                return np.log(self.W[self.tallestBin] / (w))  # log(  1/(  p/p_max  )    )


    def getFreq(self,val,timestamp=-1): #HBOS for one dimension
        bin = self.getBinIndx(val)
        self.processDecay(bin,timestamp) #if timestamp = -1, no decay will be applied
        if np.isinf(bin):
            return np.nan
        else:
            return self.W[bin]

    def getHist(self,timestamp=-1): #HBOS for one dimension
        H = np.zeros((len(self.W),1))
        for i in range(0,len(self.W)):
            self.processDecay(i,timestamp) #if timestamp = -1, no decay will be applied
            H[i] = self.W[i]
        H = H/np.sum(self.W)
        return H
    #
    # def loadFromJSON(self,jsonstring):
    #     return '' # !!!! very  important: all timestamps in self.lT should be updated so the decay won't wipe out the histogram:
    #             # self.lT = self.lT + curtime - max(self.lT)
    #             # this also applies to when the system.train setting is toggled to 'on'

from cpython cimport array

#import cython

cdef class Queue:

    cdef double[3] q
    cdef int indx
    cdef unsigned int n

    def __init__(self):
        self.q[0] = self.q[1] = self.q[2] = 0
        self.indx = 0
        self.n = 0

    cdef insert(self,double v):
        self.q[self.indx] = v
        self.indx = (self.indx + 1) % 3
        self.n += 1

    cdef unroll(self):

        cdef double[2] res
        if self.n == 2:
            res[0] = self.q[0]
            res[1] = self.q[1]
            return res
        if self.indx == 0:
            return self.q

        cdef double[3] res3
        if self.indx == 1:
            res3[0] = self.q[1]
            res3[1] = self.q[2]
            res3[2] = self.q[0]
            return res3
        else:
            res3[0] = self.q[2]
            res3[1] = self.q[0]
            res3[2] = self.q[1]
            return res3

    cdef get_last(self):
        return self.q[(self.indx-1)%3]

    cdef get_mean_diff(self):
        cdef double dif
        dif = 0
        if self.n == 2:
            dif=self.q[self.indx%3] - self.q[(self.indx-1)%3]
            return dif
        else:
            # for i in range(2):
            #     dif+=self.q[(self.indx+i+1)%3] - self.q[(self.indx+i)%3]
            dif= (self.q[self.indx%3] - self.q[(self.indx-1)%3]) + (self.q[(self.indx-1)%3] - self.q[(self.indx-2)%3])
            return dif/2

cdef class extrapolator:

    cdef Queue Qt
    cdef Queue Qv

    def __init__(self):#,int winsize=3):
        self.Qt = Queue() #deque([],winsize) #window of timestamps
        self.Qv = Queue() #deque([],winsize) #window of values

    def insert(self,double t, double v):
        self.Qt.insert(t)
        self.Qv.insert(v)

    def predict(self, double t):
        if self.Qt.n < 2: #not enough points to extrapolate?
            if self.Qt.n == 1:
                return self.Qv.get_last()
            else:
                return 0
        if (t - self.Qt.get_last())/(self.Qt.get_mean_diff() + 1e-10) > 10: # is the next timestamp 10 time further than the average sample interval?
            return self.Qv.get_last() # prediction too far ahead (very likely that we will be way off)
        cdef double yp
        cdef array.array tm = array.array('d', self.Qt.unroll())
        cdef array.array vm = array.array('d', self.Qv.unroll())
        yp = self.interpolate(t,tm,vm)
        return yp
    #TODO: try cythonize lagrange


    cdef interpolate(self, double tp, array.array tm, array.array ym):
        cdef int n
        n = len(tm) - 1
        #cdef double[:] lagrpoly = np.array([self.lagrange(tp, i, tm) for i in range(n + 1)])

        cdef double y
        for i in range(n +1):
            """
            Evaluate the i-th Lagrange polynomial at x
            based on grid data xm
            """
            y = 1
            for j in range(n + 1):
                if i != j:
                    y *= (tp - tm[j]) / (tm[i] - tm[j] + 1e-20)
            ym[i]*=y

        return sum(ym)




import numpy as np
## Prep AfterImage cython package
import os
import subprocess
import pyximport
pyximport.install()
# import after_image as af
from .after_image import IncStat1D, IncStat2D,IncStatDB
from pprint import pformat
import copy
#import AfterImage_NDSS as af

#
# MIT License
#
# Copyright (c) 2018 Yisroel mirsky
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class netStat:
    #Datastructure for efficent network stat queries
    # HostLimit: no more that this many Host identifiers will be tracked
    # HostSimplexLimit: no more that this many outgoing channels from each host will be tracked (purged periodically)
    # Lambdas: a list of 'window sizes' (decay factors) to track for each stream. nan resolved to default [5,3,1,.1,.01]
    def __init__(self, Lambdas = np.nan, HostLimit=255,HostSimplexLimit=1000, log_path=None):
        #Lambdas
        if np.isnan(Lambdas):
            self.Lambdas = [5,3,1,.1,.01]
        else:
            self.Lambdas = Lambdas
        self.clean_up_round=4000
        # number of pkts updated
        self.num_updated=0

        self.prev_pkt_time=0

        #cutoffweight for cleaning
        self.cutoffWeight=1e-6

        #HT Limits
        self.HostLimit = HostLimit
        self.SessionLimit = HostSimplexLimit*self.HostLimit*self.HostLimit #*2 since each dual creates 2 entries in memory
        self.MAC_HostLimit = self.HostLimit*10

        #HTs
        self.HT_jit = IncStatDB("HT_jit",limit=self.HostLimit*self.HostLimit)#H-H Jitter Stats
        self.HT_MI = IncStatDB("HT_MI",limit=self.MAC_HostLimit)#MAC-IP relationships
        self.HT_H = IncStatDB("HT_H",limit=self.HostLimit) #Source Host BW Stats
        self.HT_Hp = IncStatDB("HT_Hp",limit=self.SessionLimit)#Source Host BW Stats

        print("log_path:", log_path)
        if log_path is not None:
            print("netstat log_path",log_path)
            self.log_file=open(log_path,"w")
        else:
            self.log_file = None

    def set_netstat_log_path(self, log_path):
        print("netstat log_path",log_path)
        self.log_file = open(log_path,"w")

    def getHT(self):
        return {"HT_jit":self.HT_jit,
        "HT_MI" :self.HT_MI,
        "HT_H":self.HT_H,
        "HT_Hp":self.HT_Hp,
        "num_updated":self.num_updated}

    def setHT(self, HT_dict):
        self.HT_jit=HT_dict["HT_jit"]
        self.HT_MI=HT_dict["HT_MI"]
        self.HT_H=HT_dict["HT_H"]
        self.HT_Hp=HT_dict["HT_Hp"]
        self.num_updated=HT_dict["num_updated"]

    def __repr__(self):
        return "HT_jit"+pformat(self.HT_jit, indent=2)+"\nHT_MI"+pformat(self.HT_MI, indent=2)+"\nHT_H"+pformat(self.HT_H, indent=2)+"\nHT_Hp"+pformat(self.HT_Hp, indent=2)

    def findDirection(self,IPtype,srcIP,dstIP,eth_src,eth_dst): #cpp: this is all given to you in the direction string of the instance (NO NEED FOR THIS FUNCTION)
        if IPtype==0: #is IPv4
            lstP = srcIP.rfind('.')
            src_subnet = srcIP[0:lstP:]
            lstP = dstIP.rfind('.')
            dst_subnet = dstIP[0:lstP:]
        elif IPtype==1: #is IPv6
            src_subnet = srcIP[0:round(len(srcIP)/2):]
            dst_subnet = dstIP[0:round(len(dstIP)/2):]
        else: #no Network layer, use MACs
            src_subnet = eth_src
            dst_subnet = eth_dst

        return src_subnet, dst_subnet

    def get_records(self,IPtype, srcMAC,dstMAC, srcIP, srcProtocol, dstIP, dstProtocol, timestamp, datagramSize):
        db={}
        db["HT_MI"]=[]
        db["HHstat"] = [[]for i in range(len(self.Lambdas))]
        db["HHstat_cov"]=[]
        db["HHstat_jit"]=[]
        db["HpHpstat"] = [[]for i in range(len(self.Lambdas))]
        db["HpHpstat_cov"]=[]
        db["num_updated"]=self.num_updated

        for i in range(len(self.Lambdas)):
            inc_stat=self.HT_MI.register("{}_{}".format(srcMAC,srcIP), i)
            db["HT_MI"].append(inc_stat)

            inc_stat=self.HT_H.register(srcIP,i)
            db["HHstat"][i].append(inc_stat)

            inc_stat=self.HT_H.register(dstIP,i)
            db["HHstat"][i].append(inc_stat)

            inc_cov=self.HT_H.register_cov(srcIP, dstIP,i)
            db["HHstat_cov"].append(inc_cov)

            inc_stat=self.HT_jit.register("{}_{}".format(srcIP,dstIP),i,isTypeDiff=True)
            db["HHstat_jit"].append(inc_stat)


            if srcProtocol == 'arp':

                inc_stat= self.HT_Hp.register(srcMAC,i)
                db["HpHpstat"][i].append(inc_stat)

                inc_stat= self.HT_Hp.register(dstMAC,i)
                db["HpHpstat"][i].append(inc_stat)

                inc_cov=self.HT_Hp.register_cov(srcMAC, dstMAC, i)
                db["HpHpstat_cov"].append(inc_cov)
            else:  # some other protocol (e.g. TCP/UDP)
                inc_stat= self.HT_Hp.register("{}_{}".format(srcIP, srcProtocol), i)
                db["HpHpstat"][i].append(inc_stat)

                inc_stat= self.HT_Hp.register("{}_{}".format(dstIP , dstProtocol), i)
                db["HpHpstat"][i].append(inc_stat)
                inc_cov=self.HT_Hp.register_cov("{}_{}".format(srcIP, srcProtocol), "{}_{}".format(dstIP , dstProtocol),i)
                db["HpHpstat_cov"].append(inc_cov)
        return db

    def update_dummy_db(self, IPtype, srcMAC,dstMAC, srcIP, srcProtocol, dstIP, dstProtocol, timestamp, datagramSize, db,verbose=False):
        # update
        record=np.zeros(len(self.getNetStatHeaders()))

        for i in range(len(self.Lambdas)):
            if db["HT_MI"][i] is None:
                inc_stat=IncStat1D(self.Lambdas[i], "{}_{}".format(srcMAC,srcIP), timestamp)
                db["HT_MI"][i]=inc_stat
            db["HT_MI"][i].insert(datagramSize,timestamp)
            record[i*3:(i+1)*3]=db["HT_MI"][i].all_stats_1D()

            if db["HHstat"][i][0] is None:
                inc_stat=IncStat1D(self.Lambdas[i], srcIP, timestamp)
                db["HHstat"][i][0]=inc_stat
            db["HHstat"][i][0].insert(datagramSize,timestamp)

            if db["HHstat"][i][1] is None:
                inc_stat=IncStat1D(self.Lambdas[i], dstIP, timestamp)
                db["HHstat"][i][1]=inc_stat

            if db["HHstat_cov"][i] is None:
                incS1 = db["HHstat"][i][0]
                incS2 = db["HHstat"][i][1]

                inc_cov = IncStat2D(incS1,incS2,timestamp)
                incS1.add_cov(dstIP)
                incS2.add_cov(srcIP)
                db["HHstat_cov"][i]=inc_cov
            db["HHstat_cov"][i].update_cov(srcIP, datagramSize,timestamp)
            record[i*7+15:(i+1)*7+15]=db["HHstat"][i][0].all_stats_1D()+db["HHstat_cov"][i].get_stats2()



            if db["HHstat_jit"][i] is None:
                inc_stat=IncStat1D(self.Lambdas[i], "{}_{}".format(srcIP,dstIP), timestamp,isTypeDiff=True)
                db["HHstat_jit"][i]=inc_stat
            db["HHstat_jit"][i].insert(0,timestamp)
            record[(i*3)+50:((i+1)*3+50)]+=db["HHstat_jit"][i].all_stats_1D()


            if srcProtocol == 'arp':

                if db["HpHpstat"][i][0] is None:
                    inc_stat= IncStat1D(self.Lambdas[i], srcMAC, timestamp)
                    db["HpHpstat"][i][0]=inc_stat
                db["HpHpstat"][i][0].insert(datagramSize,timestamp)

                if db["HpHpstat"][i][1] is None:
                    inc_stat=IncStat1D(self.Lambdas[i], dstMAC, timestamp)
                    db["HpHpstat"][i][1]=inc_stat


                if db["HpHpstat_cov"][i] is None:
                    incS1 = db["HpHpstat"][i][0]
                    incS2 = db["HpHpstat"][i][1]

                    inc_cov = IncStat2D(incS1,incS2,timestamp)
                    incS1.add_cov(dstMAC)
                    incS2.add_cov(srcMAC)
                    db["HpHpstat_cov"][i]=inc_cov
                db["HpHpstat_cov"][i].update_cov(srcMAC,datagramSize,timestamp)

                record[(i*7)+65:((i+1)*7)+65]=db["HpHpstat"][i][0].all_stats_1D() + db["HpHpstat_cov"][i].get_stats2()
            else:

                if db["HpHpstat"][i][0] is None:
                    inc_stat=IncStat1D(self.Lambdas[i], "{}_{}".format(srcIP, srcProtocol), timestamp)
                    db["HpHpstat"][i][0] = inc_stat
                db["HpHpstat"][i][0].insert(datagramSize, timestamp)

                if db["HpHpstat"][i][1] is None:
                    inc_stat=IncStat1D(self.Lambdas[i], "{}_{}".format(dstIP , dstProtocol), timestamp)
                    db["HpHpstat"][i][1] = inc_stat

                if db["HpHpstat_cov"][i] is None:
                    incS1 = db["HpHpstat"][i][0]
                    incS2 = db["HpHpstat"][i][1]

                    inc_cov = IncStat2D(incS1,incS2,timestamp)
                    incS1.add_cov("{}_{}".format(dstIP , dstProtocol))
                    incS2.add_cov("{}_{}".format(srcIP, srcProtocol))
                    db["HpHpstat_cov"][i]=inc_cov

                db["HpHpstat_cov"][i].update_cov("{}_{}".format(srcIP, srcProtocol),datagramSize,timestamp)
                record[(i*7)+65:((i+1)*7)+65]=db["HpHpstat"][i][0].all_stats_1D() + db["HpHpstat_cov"][i].get_stats2()

        #mimic cleaning behavior
        db["num_updated"]+=1
        if db["num_updated"]%self.clean_up_round==0:
            if verbose:
                print(db["num_updated"])
                print("dummy clean timestamp",timestamp)
            self.clean_dummy_old_records(self.cutoffWeight, timestamp, db, "HT_MI", verbose)
            self.clean_dummy_old_records(self.cutoffWeight, timestamp, db, "HHstat", verbose)
            self.clean_dummy_old_records(self.cutoffWeight, timestamp, db,"HHstat_jit", verbose)
            self.clean_dummy_old_records(self.cutoffWeight, timestamp, db, "HpHpstat", verbose)
            if verbose:
                print(db)

        return record

    def clean_dummy_old_records(self,cutoffWeight,curTime,db,ht,verbose=False):
        for i in range(len(self.Lambdas)):
            if not isinstance(db[ht][i], list):
                inc_stat=db[ht][i]
                inc_stat.processDecay(curTime)
                if inc_stat.weight() < cutoffWeight:
                    # print("cleaning dummy record",int_stat)
                    db[ht][i]=None
            else:
                inc_stat=db[ht][i][0]
                inc_stat.processDecay(curTime)
                inc_stat2=db[ht][i][1]
                inc_stat2.processDecay(curTime)

                remove_flag=False
                if inc_stat.weight() < cutoffWeight:
                    if verbose:
                        print("dummy inc stat", inc_stat)
                    db[ht][i][1].remove_cov(inc_stat.name)
                    db[ht][i][0]=None

                    remove_flag=True

                if inc_stat2.weight()<cutoffWeight:
                    if verbose:
                        print("dummy inc stat", inc_stat2)
                    db[ht][i][0].remove_cov(inc_stat2.name)
                    db[ht][i][1]=None

                    remove_flag=True

                if remove_flag:
                    db[ht+"_cov"][i]=None

    def updateGetStats(self, IPtype, srcMAC,dstMAC, srcIP, srcProtocol, dstIP, dstProtocol, timestamp, datagramSize):
        # Host BW: Stats on the srcIP's general Sender Statistics
        # Hstat = np.zeros((3*len(self.Lambdas,)))
        # for i in range(len(self.Lambdas)):
        #     Hstat[(i*3):((i+1)*3)] = self.HT_H.update_get_1D_Stats(srcIP, timestamp, datagramSize, self.Lambdas[i])

        #MAC.IP: Stats on src MAC-IP relationships
        if self.log_file is not None:
            self.log_file.write("{}, {}, {}, {}, {}, {}, {}, {}, {}\n".format( IPtype, srcMAC,dstMAC, srcIP, srcProtocol, dstIP, dstProtocol, timestamp, datagramSize))

        self.prev_pkt_time=timestamp
        record=np.zeros(len(self.getNetStatHeaders()))

        for i in range(len(self.Lambdas)):
            record[i*3:(i+1)*3]=self.HT_MI.update_get_1D_Stats("{}_{}".format(srcMAC,srcIP), timestamp, datagramSize, i)
            record[i*7+15:(i+1)*7+15]=self.HT_H.update_get_1D2D_Stats(srcIP, dstIP,timestamp,datagramSize,i)
            record[(i*3)+50:((i+1)*3+50)]=self.HT_jit.update_get_1D_Stats("{}_{}".format(srcIP,dstIP), timestamp, 0,i,isTypeDiff=True)
            if srcProtocol == 'arp':
                record[(i*7)+65:((i+1)*7)+65]=self.HT_Hp.update_get_1D2D_Stats(srcMAC, dstMAC, timestamp, datagramSize, i)
            else:
                record[(i*7)+65:((i+1)*7)+65] = self.HT_Hp.update_get_1D2D_Stats("{}_{}".format(srcIP, srcProtocol), "{}_{}".format(dstIP , dstProtocol), timestamp, datagramSize, i)

        #
        # MIstat =  np.zeros((3*len(self.Lambdas,)))
        # for i in range(len(self.Lambdas)):
        #     MIstat[(i*3):((i+1)*3)] = self.HT_MI.update_get_1D_Stats("{}_{}".format(srcMAC,srcIP), timestamp, datagramSize, i)
        #
        # # Host-Host BW: Stats on the dual traffic behavior between srcIP and dstIP
        # HHstat =  np.zeros((7*len(self.Lambdas,)))
        # for i in range(len(self.Lambdas)):
        #     HHstat[(i*7):((i+1)*7)] = self.HT_H.update_get_1D2D_Stats(srcIP, dstIP,timestamp,datagramSize,i)
        # # Host-Host Jitter:
        # HHstat_jit =  np.zeros((3*len(self.Lambdas,)))
        # for i in range(len(self.Lambdas)):
        #     HHstat_jit[(i*3):((i+1)*3)] = self.HT_jit.update_get_1D_Stats("{}_{}".format(srcIP,dstIP), timestamp, 0,i,isTypeDiff=True)
        #
        # # Host-Host BW: Stats on the dual traffic behavior between srcIP and dstIP
        # HpHpstat =  np.zeros((7*len(self.Lambdas,)))
        # if srcProtocol == 'arp':
        #     for i in range(len(self.Lambdas)):
        #         HpHpstat[(i*7):((i+1)*7)] = self.HT_Hp.update_get_1D2D_Stats(srcMAC, dstMAC, timestamp, datagramSize, i)
        # else:  # some other protocol (e.g. TCP/UDP)
        #     for i in range(len(self.Lambdas)):
        #         HpHpstat[(i*7):((i+1)*7)] = self.HT_Hp.update_get_1D2D_Stats("{}_{}".format(srcIP, srcProtocol), "{}_{}".format(dstIP , dstProtocol), timestamp, datagramSize, i)

        self.num_updated+=1

        #clean our records every 100 updates
        if self.num_updated%self.clean_up_round==0:

            self.HT_MI.cleanOutOldRecords(self.cutoffWeight, timestamp)
            self.HT_H.cleanOutOldRecords(self.cutoffWeight, timestamp)
            self.HT_jit.cleanOutOldRecords(self.cutoffWeight, timestamp)
            self.HT_Hp.cleanOutOldRecords(self.cutoffWeight, timestamp)
            # print(c1,c2,c3,c4)
            # print(n1,n2,n3,n4, self.cutoffWeight)
            # print(self.HT_Hp.stat1d[4].keys())

        return record

    def get_stats(self, IPtype, srcMAC, dstMAC, srcIP, srcProtocol, dstIP, dstProtocol, t1, frame_len=None, Lambda=1):
        """get stats of a packet, framelen not needed"""

        MIstat =  np.zeros((3*len(self.Lambdas,)))
        for i in range(len(self.Lambdas)):
            ID=srcMAC+srcIP
            MIstat[(i*3):((i+1)*3)] = self.HT_MI.get_1D_Stats(ID,self.Lambdas[i])

        HHstat =  np.zeros((7*len(self.Lambdas,)))
        for i in range(len(self.Lambdas)):
            HHstat[(i*7):((i+1)*7)] = self.HT_H.get_1D_Stats(srcIP,self.Lambdas[i])+self.HT_H.get_2D_stats(srcIP, dstIP,self.Lambdas[i],t1,level=2)


        # amount of traffic arriving out of time order
        HHstat_jit =  np.zeros((3*len(self.Lambdas,)))
        for i in range(len(self.Lambdas)):
            HHstat_jit[(i*3):((i+1)*3)] = self.HT_jit.get_1D_Stats(srcIP+dstIP,self.Lambdas[i])

        HpHpstat =  np.zeros((7*len(self.Lambdas,)))
        if srcProtocol == 'arp':
            for i in range(len(self.Lambdas)):
                HpHpstat[(i*7):((i+1)*7)] = self.HT_Hp.get_1D_Stats(srcMAC,self.Lambdas[i])+self.HT_Hp.get_2D_stats(srcMAC, dstMAC,self.Lambdas[i],t1,level=2)
        else:  # some other protocol (e.g. TCP/UDP)
            for i in range(len(self.Lambdas)):
                HpHpstat[(i*7):((i+1)*7)] = self.HT_Hp.get_1D_Stats(srcIP + srcProtocol, self.Lambdas[i])+self.HT_Hp.get_2D_stats(srcIP + srcProtocol, dstIP + dstProtocol,self.Lambdas[i],t1,level=2)

        return np.concatenate((MIstat,HHstat, HHstat_jit,HpHpstat))


    def getNetStatHeaders(self):

        MIstat_headers = self.HT_MI.get_headers()
        HHstat_headers = self.HT_H.get_headers(True)
        HHjitstat_headers = self.HT_jit.get_headers()
        HpHpstat_headers = self.HT_Hp.get_headers(True)


        return MIstat_headers + HHstat_headers + HHjitstat_headers + HpHpstat_headers



import numpy as np
from pprint import pformat


def magnitude(mean, other_mean):
    # the magnitude of a set of incStats, pass var instead of mean to get radius
    return np.sqrt(np.power(mean,2)+np.power(other_mean,2))

class IncStat1D:
    def __init__(self, l, name, init_time=0, isTypeDiff=False):  # timestamp is creation time
        self.name = name
        self.ls = 0  # linear sum
        self.ss = 0  # sum of squares
        self.w = 1e-20  # weight
        self.weight_thresh=1e-6
        self.isTypeDiff = isTypeDiff
        self.l = l  # Decay Factor
        self.lastTimestamp = init_time
        self.covs = [] # a list of incStat_covs (references) with relate to this incStat

    def __repr__(self):
        return pformat(vars(self))

    def add_cov(self, ID):
        self.covs.append(ID)

    def remove_cov(self, ID):
        self.covs.remove(ID)


    def insert(self, v, t=0):  # v is a scalar, t is v's arrival the timestamp
        # print("last t", self.lastTimestamp)
        # isTypeDiff= traffic jitter
        if self.isTypeDiff:
            dif = t - self.lastTimestamp
            if dif > 0:
                v = dif
            else:
                v = 0

        self.processDecay(t)

        # update with v
        self.ls += v
        self.ss += np.power(v,2)
        self.w += 1

    def processDecay(self, timestamp):
        # check for decay
        timeDiff = timestamp - self.lastTimestamp
        if timeDiff > 0:
            factor = np.power(2,  (-self.l * timeDiff))

            self.ls = self.ls * factor
            self.ss = self.ss * factor
            self.w = self.w * factor
            self.lastTimestamp = timestamp

    def weight(self):
        return self.w

    def mean(self):
        if self.w < self.weight_thresh:
            return 0
        return self.ls / self.w

    def var(self):
        if self.w < self.weight_thresh:
            return 0
        return max(0., self.ss / self.w - np.power(self.mean(), 2))

    def std(self):
        return np.sqrt(self.var())

    def all_stats_1D(self):
        return [self.weight(), self.mean(), self.var()]

class IncStat2D:
    def __init__(self, incS1, incS2, init_time = 0):
        # store references tot he streams' incStats
        self.incStats = [incS1,incS2]
        self.lastRes = [0,0]
        # init extrapolators
        #self.EXs = [extrapolator(),extrapolator()]

        # init sum product residuals
        self.sr = 0 # sum of residule products (A-uA)(B-uB)
        self.lastTimestamp = init_time

        self.w3=0


    def __repr__(self):
        return "{}, {} : ".format(self.incStats[0].name,self.incStats[1].name)+pformat(self.get_stats2())

    #other_incS_decay is the decay factor of the other incstat
    # ID: the stream ID which produced (v,t)
    def update_cov(self, ID, v, t):  # it is assumes that incStat "ID" has ALREADY been updated with (t,v) [this si performed automatically in method incStat.insert()]
        # find incStat
        if ID == self.incStats[0].name:
            inc = 0
        elif ID == self.incStats[1].name:
            inc = 1
        else:
            print("update_cov ID error:", ID)
            return ## error

        # Decay other incStat, assuming this incstat is already decayed in 1d
        self.incStats[not(inc)].processDecay(t)

        # Decay residules
        self.processDecay(t,inc)

        # Update extrapolator for current stream
        #self.EXs[inc].insert(t,v)

        # Extrapolate other stream
        #v_other = self.EXs[not(inc)].predict(t)

        # Compute and update residule
        res = (v - self.incStats[inc].mean())
        resid = res * self.lastRes[not(inc)]
        self.sr += resid
        self.lastRes[inc] = res
        self.w3 += 1


    def processDecay(self,t,i):
        # check for decay cf3
        timeDiffs = t - self.lastTimestamp
        if timeDiffs > 0:
            factor = np.power(2, (-self.incStats[i].l * timeDiffs))
            self.sr *= factor
            self.lastTimestamp = t
            self.w3 *= factor
            self.lastRes[i] *= factor

    #todo: add W3 for cf3
    def get_stats2(self):
        mean1=self.incStats[0].mean()
        mean2=self.incStats[1].mean()
        var1=self.incStats[0].var()
        var2=self.incStats[1].var()


        return [magnitude(var1,var2),magnitude(mean1,mean2),self.cov(), self.pcc()]


    #covariance approximation
    def cov(self):
        return self.sr / (self.incStats[0].weight()+self.incStats[1].weight())
        # return self.sr / self.w3

    # Pearson corl. coef
    def pcc(self):
        ss = self.incStats[0].std() * self.incStats[1].std()
        if ss != 0:
            return self.cov() / ss
        else:
            return 0


class IncStatDB:
    # default_lambda: use this as the lambda for all streams. If not specified, then you must supply a Lambda with every query.
    def __init__(self, name, limit=1e5, lambdas=[5,3,1,.1,.01]):
        # list of dictionary to store 1d stats for each lambda, index matches lambda id
        self.stat1d = [ {} for i in range(len(lambdas))]
        # list of dict to store 2d stats for each lambda, index matches lambda id
        self.stat2d=[ {} for i in range(len(lambdas))]
        # limit for all lambdas combined
        self.limit = limit
        self.lambdas = lambdas
        self.num_entries=0
        self.name=name


    def get_headers(self, include_2d=False):
        fmt_str=self.name+"_{}_{}"
        stat_headers=self.get_1d_headers()
        if include_2d:
            stat_headers+=self.get_2d_headers()

        header=[fmt_str.format(l,n) for l in self.lambdas for n in stat_headers]

        return header

    # Registers a new stream. init_time: init lastTimestamp of the incStat
    def register(self,ID,lambda_index,init_time=None,isTypeDiff=False):
        # not in our db
        key = ID+"_"+str(self.lambdas)
        if ID not in self.stat1d[lambda_index]:
            if init_time is None:
                return None
            if self.num_entries + 1 > self.limit:
                raise LookupError(
                    'Adding Entry:\n' + key + '\nwould exceed incStat 1D limit of ' + str(
                        self.limit) + '.\nObservation Rejected.')
            stat1d = IncStat1D(self.lambdas[lambda_index], ID, init_time, isTypeDiff)

            self.stat1d[lambda_index][ID] = stat1d

            self.num_entries+=1
        else:
            stat1d = self.stat1d[lambda_index].get(ID)
        return stat1d

    # updates/registers stream
    def update(self,ID,t,v,lambda_index,isTypeDiff=False):
        stat_1d = self.register(ID,lambda_index,t,isTypeDiff)
        stat_1d.insert(v,t)
        return stat_1d

    # Updates and then pulls current 1D stats from the given ID. Automatically registers previously unknown stream IDs
    def update_get_1D_Stats(self, ID, t, v, lambda_index, isTypeDiff=False):  # weight, mean, std
        incS = self.update(ID,t,v,lambda_index,isTypeDiff)
        return incS.all_stats_1D()

    def update_get_1D2D_Stats(self, ID1,ID2,t1,v1,lambda_index):  # weight, mean, std
        return self.update_get_1D_Stats(ID1,t1,v1,lambda_index) + self.update_get_2D_Stats(ID1,ID2,t1,v1,lambda_index)


    def update_get_2D_Stats(self, ID1,ID2,t1,v1,lambda_index):  #level=  1:cov,pcc  2:radius,magnitude,cov,pcc
        #retrieve/add cov tracker
        inc_cov = self.register_cov(ID1, ID2, lambda_index,  t1)
        # Update cov tracker
        inc_cov.update_cov(ID1,v1,t1)
        return inc_cov.get_stats2()

    # Registers covariance tracking for two streams, registers missing streams
    def register_cov(self,ID1,ID2,lambda_index,init_time=None,isTypeDiff=False):

        # Lookup both streams
        incS1 = self.register(ID1,lambda_index,init_time,isTypeDiff)
        incS2 = self.register(ID2,lambda_index,init_time,isTypeDiff)

        #check for pre-exiting link
        if (ID1, ID2) in self.stat2d[lambda_index]:
            return self.stat2d[lambda_index][(ID1,ID2)]
        elif (ID2, ID1) in self.stat2d[lambda_index]:
            return self.stat2d[lambda_index][(ID2,ID1)]

        else:
            # Link incStats
            if init_time is None:
                return None
            inc_cov = IncStat2D(incS1,incS2,init_time)
            self.stat2d[lambda_index][(ID1,ID2)]=inc_cov
            incS1.add_cov(ID2)
            incS2.add_cov(ID1)
            return inc_cov


    def get_1d_headers(self):
        return ["weight","mean","std"]

    def get_2d_headers(self):
        return ["radius","magnitude","covariance","pcc"]


    #cleans out records that have a weight less than the cutoff.
    #returns number of removed records and records looked through
    def cleanOutOldRecords(self, cutoffWeight, curTime, verbose=False):
        n = 0
        cleaned=0
        for i in range(len(self.lambdas)):
            for key, inc_stat in dict(self.stat1d[i]).items():
                inc_stat.processDecay(curTime)
                cleaned+=1
                if inc_stat.weight() < cutoffWeight:
                    # remove all links
                    for other_is in inc_stat.covs:
                        try:
                            del self.stat2d[i][(inc_stat.name, other_is)]
                        except KeyError:
                            del self.stat2d[i][(other_is, inc_stat.name)]
                        self.stat1d[i][other_is].remove_cov(inc_stat.name)
                    # remove 1d
                    del self.stat1d[i][key]
                    n+=1
        return n, cleaned



from doctest import UnexpectedException
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import torch
import torch.nn as nn
from scipy.cluster.hierarchy import ClusterNode, linkage, to_tree


class FeatureMapper(object):
    """Computes a hierarchical cluster on the given dataset columns.

    Parameters
    ----------
    dim : int
        Number of dataset columns (or features).
    max_features_per_cluster : int
        Maximum numbers of features in a cluster.
    """

    def __init__(self, dim: int, max_features_per_cluster: int) -> None:
        # Algorithm hyperparameters
        self.dim = dim
        self.max_features_per_cluster = max_features_per_cluster

        # Internal state
        self.running_sum: torch.Tensor
        self.running_residuals: torch.Tensor
        self.running_squared_residuals: torch.Tensor
        self.running_corr_mat: torch.Tensor
        self.seen_samples: int

        # Variable to lazily evaluate the clusters_ property
        self._clusters = []

        # Check if fit method has already been called
        self._fitted: bool = False

        self._init_stats()

    def _init_stats(self) -> None:
        # linear num of features
        self.running_sum = torch.zeros(self.dim,
                                       requires_grad=False,
                                       dtype=torch.float64)

        # linear sum of feature residuals
        self.running_residuals = torch.zeros(self.dim,
                                             requires_grad=False,
                                             dtype=torch.float64)

        # squared sum of feature residuals
        self.running_squared_residuals = torch.zeros(self.dim,
                                                     requires_grad=False,
                                                     dtype=torch.float64)

        # partial correlation matrix
        self.running_corr_mat = torch.zeros((self.dim, self.dim),
                                            requires_grad=False,
                                            dtype=torch.float64)
        self.seen_samples = 0
        self._fitted = False
        self._clusters = []

    def parital_fit(self, x: torch.Tensor) -> None:
        """Updates the internal statistics based on data batch.

        Parameters
        ----------
        x : torch.Tensor
            Dataset batch.

        Raises
        ------
        ValueError
            In case the number of columns is not the same as the one specified
            in the init. Otherwise this is triggered if the given x has an
            invalid number of dimensions.
        """
        if x.dim() not in {1, 2}:
            raise ValueError(f"Invalid number of input axes: {x.dim()}")

        if x.size(-1) != self.dim:
            raise ValueError(f"Invalid number of input dims: got {x.size(-1)} "
                             f"expected {self.dim}")

        if x.dim() == 1:
            x = x.unsqueeze(0)

        self._update_stats(x)
        self._update_clusters()
        self._fitted = True

    def fit(self, x: torch.Tensor) -> None:
        """Updates the internal statistics with the whole dataset at once.

        Parameters
        ----------
        x : torch.Tensor
            Processed dataset.
        """
        self._init_stats()
        self.parital_fit(x)

    def transform(self, x: torch.Tensor) -> List[torch.Tensor]:
        """Splits the given data into subgroups.

        Each group is a subset of columns of the given data based on the
        clustering strategy learned on the `fit` method.

        Parameters
        ----------
        x : torch.Tensor
            Data to transform

        Returns
        -------
        List[torch.Tensor]
            Groups of columns.

        Raises
        ------
        ValueError
            If the x dimensions is invalid.
        """
        if x.dim() not in {1, 2}:
            raise ValueError(f"Invalid number of input axes: {x.dim()}")

        if x.dim() == 1:
            x = x.unsqueeze(0)

        return [x[:, torch.as_tensor(indices)] for indices in self.clusters_]

    @property
    def clusters_(self) -> List[List[int]]:
        """Property to get the learned clusters.

        Returns
        -------
        List[List[int]]
            Each position of the list contains a cluster with the corresponding
            column indices.

        Raises
        ------
        UnexpectedException
            If `fit` or `partial_fit` has not been called yet.
        """
        if not self._fitted:
            raise UnexpectedException(
                "fit or partial_fit has not been called yet")
        return self._clusters

    @property
    def n_clusters_(self) -> int:
        return len(self.clusters_)

    def _update_stats(self, x: torch.Tensor) -> None:
        self.seen_samples += x.size(0)
        self.running_sum += x.sum(0)
        running_mean = self.running_sum / float(self.seen_samples)
        residuals = x - running_mean
        self.running_residuals += residuals.sum(0)
        self.running_squared_residuals += residuals.pow(2.0).sum(0)
        self.running_corr_mat += sum(torch.outer(r, r) for r in residuals)

    def _update_clusters(self) -> None:
        dist = self._compute_correlation_distance()
        indices = torch.triu_indices(self.dim, self.dim, offset=1)
        link_mat = linkage(dist[indices[0], indices[1]].cpu().numpy())
        self._clusters = self._split_clusters(to_tree(link_mat))
        self._clusters = self._join_clusters(self._clusters)

    def _compute_correlation_distance(self) -> torch.Tensor:
        rs_sqrt = self.running_squared_residuals.sqrt()
        rs_sqrt_mat = torch.outer(rs_sqrt, rs_sqrt)
        rs_sqrt_mat[rs_sqrt_mat.eq(0)] = 1e-16
        dist = 1.0 - self.running_corr_mat / rs_sqrt_mat
        dist[dist.lt(0)] = 0.0
        return dist

    def _join_clusters(self, clusters: List[List[int]]) -> List[List[int]]:
        # Join clusters until they at least have n elements
        # This not follows any heuristic, it just merges clusters to meet
        # the required cardinality (TODO improve this)
        while True:
            ci = [i for i, o in enumerate(clusters) if len(o) < 4]
            if not ci:
                return clusters
            elif len(ci) == 1:
                c = clusters.pop(ci[0])
                min(clusters, key=len).extend(c)
            else:
                merged_cluster = clusters.pop(ci[0]) + clusters.pop(ci[1] - 1)
                clusters.append(merged_cluster)

    def _split_clusters(self, cn: ClusterNode) -> List[List[int]]:
        # If clusters are to large split them recursively
        if cn.get_count() <= self.max_features_per_cluster:
            return [cn.pre_order()]

        left_clusters = self._split_clusters(cn.get_left())
        right_clusters = self._split_clusters(cn.get_right())
        return left_clusters + right_clusters


class TinyAutoEncoder(nn.Module):
    """AutoEncoder with one compresion layer.

    The inputs must be normalized between 0 and 1. For instance you can use
    a MinMaxScaler from scikit-learn.

    Parameters
    ----------
    in_features : int
        Number of input features. The reconstructed output will have the same
        output.
    compression_rate : float
        Percentage to compress the hidden auto encoder representation space.
        For example if `in_features=100` and `compression_rate=0.6` then the 
        hidden dimension will have 40 dimensions. Defaults 0.6.
    dropout_rate : float
        Dropout rate applied to input features. Defaults 0.2.
    """

    def __init__(self,
                 *,
                 in_features: int,
                 compression_rate: float = 0.6,
                 dropout_rate: float = 0.2) -> None:
        super().__init__()
        self.in_features = in_features
        self.compression_rate = compression_rate
        self.dropout_rate = dropout_rate

        hidden_units = int(in_features * (1.0 - self.compression_rate))

        # Initialize weight
        w_data = torch.empty(in_features, hidden_units)
        w_data.uniform_(-1. / in_features, to=1. / in_features)
        self.w = nn.Parameter(w_data)

        # Initialize biases
        self.hidden_bias = nn.Parameter(torch.zeros(hidden_units))
        self.reconstruct_bias = nn.Parameter(torch.zeros(in_features))

        self.dropout = nn.Dropout(p=self.dropout_rate)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.dropout(x)
        hidden = (x @ self.w + self.hidden_bias).tanh()
        return (hidden @ self.w.t() + self.reconstruct_bias).sigmoid()


class RMSELoss(torch.nn.Module):
    """Convinen Root Mean Squared Error implementation.

    Just applies the `sqrt` method to the vanilla PyTorch MSELoss.
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__()
        self.criterion = nn.MSELoss(**kwargs)

    def forward(self, inputs: torch.Tensor,
                targets: torch.Tensor) -> torch.Tensor:
        return torch.sqrt(self.criterion(inputs, targets) + 1e-7)


class Kitsune(nn.Module):

    def __init__(self,
                 feature_mapper: FeatureMapper,
                 compression_rate: float = 0.6,
                 dropout_rate: float = 0.1) -> None:
        super().__init__()
        self.fm = feature_mapper
        if not self.fm._fitted:
            raise ValueError("Kitsune needs a fitted `feature_map`. "
                             "Make sure you call `.fit` method before hand.")

        self.compression_rate = compression_rate
        self.dropout_rate = dropout_rate

        self.tails = self._build_tails(compression_rate, dropout_rate)
        self.head = TinyAutoEncoder(in_features=self.fm.n_clusters_,
                                    compression_rate=compression_rate,
                                    dropout_rate=dropout_rate)

        self.mse = RMSELoss(reduction="none")

    def _build_tails(self, compression_rate: float,
                     dropout_rate: float) -> List[nn.Module]:
        return nn.ModuleList([
            TinyAutoEncoder(in_features=len(c),
                            compression_rate=compression_rate,
                            dropout_rate=dropout_rate)
            for c in self.fm.clusters_
        ])

    def forward(self, x: torch.Tensor) -> Dict[str, torch.Tensor]:
        split_feat = self.fm.transform(x)
        tails_losses = torch.empty(x.size(0), len(self.tails), device=x.device)
        for i, (feat, tail) in enumerate(zip(split_feat, self.tails)):
            reconstructed_feat = tail(feat)
            tails_losses[:, i] = self.mse(reconstructed_feat, feat).mean(-1)

        reconstructed_loss_dist = self.head(tails_losses.detach())
        head_loss = self.mse(reconstructed_loss_dist, tails_losses.detach())
        return {
            "tails_losses": tails_losses,
            "head_loss": head_loss.mean(-1),
        }

    def score(self, x: torch.Tensor) -> torch.Tensor:
        return self(x)["head_loss"]

    def save(self, fpath: Union[str, Path]) -> None:
        chkp = {
            "config": {
                "feature_mapper": self.fm,
                "compression_rate": self.compression_rate,
                "dropout_rate": self.dropout_rate,
            },
            "weights": self.state_dict(),
        }
        torch.save(chkp, fpath)

    @classmethod
    def from_pretrained(cls,
                        fpath: Union[str, Path],
                        map_location: Optional[torch.device] = None) -> None:
        chkp = torch.load(fpath, map_location=map_location)
        model = cls(**chkp.pop("config"))
        model.eval()
        model.load_state_dict(chkp.pop("weights"))
        return model



"""
# author: Swain Subrat Kumar
A demo __init__.py file should contain all 
the public API that need to be exported 
from this package
"""

from .awesome_ids_model import AwesomeIDS

__all__ = [
    "AwesomeIDS"
]



import random
from pathlib import Path

import numpy as np
import torch

from .data import FileFormat, build_input_data_pipe
from .engine import build_feature_mapper, train_single_epoch
from .models import Kitsune

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def train(
    input_path: Path,
    batch_size: int = 32,
    file_format: FileFormat = "csv",
    compression_rate: float = 0.6,
    checkpoint_dir: Path = Path("models/")) -> None:

    random.seed(0)
    torch.manual_seed(0)
    np.random.seed(0)

    dp = build_input_data_pipe(str(input_path),
                               batch_size=batch_size,
                               shuffle=True,
                               file_format=file_format)

    print("🦊 Training feature mapper ...")
    feature_mapper = build_feature_mapper(dp,
                                          ds_features=115,
                                          max_features_per_cluster=10)

    model = Kitsune(feature_mapper=feature_mapper,
                    compression_rate=compression_rate)
    model.to(device)

    optimizer = torch.optim.SGD(model.parameters(), lr=1e-3, momentum=0.9)

    print("🦊 Training Kitsune ensemble ...")
    train_single_epoch(model=model, ds=dp, optimizer=optimizer, device=device)

    print(f"🦊 Serializing the model to {checkpoint_dir / 'kitsune.pt'} ...")
    model.save(checkpoint_dir / "kitsune.pt")



import typer

from kitsune.anomaly_score import compute_anomaly_scores
from kitsune.train import train

kitsune_cli = typer.Typer(name="Kitsune Network")
kitsune_cli.command(name="train")(train)
kitsune_cli.command(name="anomaly-scores")(compute_anomaly_scores)


@kitsune_cli.callback()
def main() -> None:
    """Kitsune anomaly detector Command Line Interface."""


if __name__ == "__main__":
    kitsune_cli()



from typing import Iterable

import torch
import torch.nn as nn

from .models import FeatureMapper, Kitsune


def build_feature_mapper(ds: Iterable[torch.Tensor],
                         ds_features: int,
                         max_features_per_cluster: int = 10) -> FeatureMapper:
    fm = FeatureMapper(ds_features, max_features_per_cluster)
    for batch in ds:
        fm.parital_fit(batch)
    return fm


@torch.inference_mode()
def predict(model: Kitsune, 
            ds: Iterable[torch.Tensor], 
            device: torch.device = torch.device("cpu")) -> torch.Tensor:
    predictions = []
    model.eval()
    for batch in ds:
        batch = batch.to(device)
        predictions.extend(model.score(batch).cpu().tolist())
    return torch.as_tensor(predictions)


def train_single_epoch(
    model: Kitsune,
    ds: Iterable[torch.Tensor],
    optimizer: torch.optim.Optimizer,
    log_every: int = 60,
    epoch: int = 0,
    device: torch.device = torch.device("cpu")) -> None:
    """Train one round using all the data.

    Parameters
    ----------
    v : Kitsune
        Kitsune model to train
    ds : Iterable[torch.Tensor]
        Itarable to fetch the processed data. Usually a DataLoader
    optimizer : torch.optim.Optimizer
        PyTorch optimizer to optimize the kitsune parameters
    log_every : int, optional
        Report the loss every n steps, by default 30
    epoch : int, optional
        Only for log purposes, current epoch, by default 0
    device : torch.device, optional
        Device where the model is located, by default torch.device("cpu")
    """
    model.to(device)
    model.train()
    model.zero_grad()

    running_tail = 0.0
    running_head = 0.0

    for i, sample in enumerate(ds, start=1):
        sample = sample.to(device)

        optimizer.zero_grad()
        losses = model(sample.float())
        tail_loss = losses["tails_losses"].sum(-1).mean(0)
        head_loss = losses["head_loss"].mean(0)
        tail_loss.backward()
        head_loss.backward()
        nn.utils.clip_grad.clip_grad_norm_(model.parameters(), 2.0)
        optimizer.step()

        running_tail += tail_loss.item()
        running_head += head_loss.item()

        if i % log_every == 0:
            tail_loss_mean = running_tail / i
            head_loss_mean = running_head / i
            print(f"🦊 Epoch [{epoch}]  "
                  f"tail losses {tail_loss_mean:.5f}  "
                  f"head loss: {head_loss_mean:.5f}")



# extract features
from .FeatureExtractor import FE
from .transformation import Transformation, generate_key
import os

#pcap file to features tsv using wireshark
def packet_parser(pcap_file, output_file):
    command = "tshark -r {} -T fields -e frame.time_epoch -e frame.len -e frame.protocols -e ip.src -e ip.dst -e ip.proto -e tcp.srcport -e tcp.dstport -e udp.srcport -e udp.dstport -e dns.qry.name -e dns.qry.type -e dns.resp.name -e dns.resp.type -e dns.flags.response -e dns.flags.rcode -e dns.flags.recursion_desired -e dns.flags.recursion_available -e dns.flags.check_disable -e dns.flags.authentic_data -e dns.flags.truncated -e dns.flags.z -e dns.flags.authentic_data -e dns.flags.check_disable -e dns.flags.recursion_available -e dns.flags.recursion_desired -e dns.flags.rcode -e dns.flags.response -e dns.resp.type -e dns.resp.name -e dns.qry.type -e dns.qry.name -e udp.dstport -e udp.srcport -e tcp.dstport -e tcp.srcport -e ip.proto -e ip.dst -e ip.src -e frame.protocols -e frame.len -e frame.time_epoch > {}".format(pcap_file, output_file)
    os.system(command)

def extract_features(data_path, output_path):
    max_vec = [-1]*100
    min_vec = [1e10]*100
    fe = FE(data_path)   
    while True:
        fv = fe.get_next_vector()
        if len(fv) == 0:
            break
        for i in range(len(fv)):
            if fv[i] > max_vec[i]:
                max_vec[i] = fv[i]
            if fv[i] < min_vec[i]:
                min_vec[i] = fv[i]

    fe = FE(data_path)
    f = open(output_path, "w")
    while True:
        fv = fe.get_next_vector()
        if len(fv) == 0:
            break
        for i in range(len(fv)):
            fv[i] = (fv[i] - min_vec[i]) / (max_vec[i] - min_vec[i])
        f.write(",".join([str(x) for x in fv]) + "\n")

    f.close()



def transform_features(data_path, output_path):
    inputf = open(data_path, "r")
    outputf = open(output_path, "w")

    key = generate_key(64)
    T = Transformation("1", key)
    
    while True:
        line = inputf.readline()
        if len(line) == 0:
            break
        fv = [float(x) for x in line.split(",")]
        fv = T.transform(fv)
        outputf.write(",".join([str(x) for x in fv]) + "\n")

    inputf.close()
    outputf.close()
    




import time
from itertools import product
from .after_image.feature_extractor import *
from tqdm import tqdm
import multiprocessing as mp
import pickle

np.set_printoptions(suppress=True,
                    formatter={'float_kind': '{:f}'.format})


def parse_kitsune(pcap_file, output_file_name, add_label=False, write_prob=1, count=float('Inf'), parse_type="scapy", netstat_path=None, save_netstat=None, add_proto=False, add_time=False, netstat_log_file=None):
    """Short summary.

    Args:
        pcap_file (string): path to pcap file..
        output_file_name (string): output path of the feature file. .
        add_label (boolean): whether to add label at the end of feature. if true, label is the filename. Defaults to False.
        write_prob (float 0~1): probability of writing a feature to file. 1 indicates all features are written. Defaults to 1.
        count (int): number of packets to process. Defaults to float('Inf').
        parse_type (string): either scapy or tshark. There are some bugs with tshark parsing so stick with scapy. Defaults to "scapy".

    Returns:
        nothing, file is written to output_file_name

    Raises:
        EOFError,ValueError,StopIteration: when EOF is reached. already handled

    """
    print("parsing:", pcap_file)

    nstat = None
    if netstat_path:
        print("loading netstat from", netstat_path)
        with open(netstat_path, "rb") as m:
            nstat = pickle.load(m)

    feature_extractor = FE(pcap_file, parse_type=parse_type,
                           nstat=nstat, log_file=netstat_log_file)
    # temp=open("tmp.txt","w")
    headers = feature_extractor.nstat.getNetStatHeaders()
    output_file = open(output_file_name, "w")
    label = output_file_name.split('/')[-1]
    if add_label:
        headers += ["label"]

    if add_proto:
        headers += ["protocol"]

    if add_time:
        headers += ["time"]

    # print(headers)
    np.savetxt(output_file, [headers], fmt="%s", delimiter=",")

    skipped = 0
    written = 0
    t = tqdm(total=count)
    pkt_index = 0
    while pkt_index < count:
        try:
            if parse_type == "scapy":
                traffic_data, packet = feature_extractor.get_next_vector()
            else:
                traffic_data = feature_extractor.get_next_vector()
        except EOFError as e:
            print("EOF Reached")
            print(e)
            break
        except ValueError as e:
            print("EOF Reached")
            print(e)
            break
        except StopIteration as e:
            print(e)
            print("EOF Reached")
            break

        pkt_index += 1
        t.update(1)
        if traffic_data == []:
            np.savetxt(output_file, np.full(
                features.shape, -1), delimiter=",")
            # print(pkt_index)
            skipped += 1
            continue
        # print(traffic_data)
        features = feature_extractor.nstat.updateGetStats(*traffic_data)
        # protocol = traffic_data[4]

        if np.isnan(features).any():
            print(features)
            break
        # temp.write("{}\n".format(pkt_index))
        if np.random.uniform(0, 1) < write_prob:

            if add_label:
                features = np.append(features, label)
            if add_proto:

                layers = packet.layers()
                while packet[layers[-1]].name in ["Raw", "Padding"]:
                    del layers[-1]
                protocol = packet[layers[-1]].name
                features = np.append(features, protocol)
            if add_time:
                time = traffic_data[-2]
                features = np.append(features, time)
            features = np.expand_dims(features, axis=0)
            np.savetxt(output_file, features, delimiter=",", fmt="%s")

        written += 1
    t.close()

    if save_netstat:
        with open(save_netstat, "wb") as out:
            pickle.dump(feature_extractor.get_nstat(), out)

    output_file.close()
    print("skipped:", skipped)
    print("written:", written)



#Check if cython code has been compiled
import os
import subprocess
import copy
#Import dependencies
from .net_stat import netStat
import csv
import numpy as np
print("Importing Scapy Library")
from scapy.all import *
import os.path
import platform
import subprocess
import pickle
from scapy.layers.http import *
from scapy.layers.inet import TCP


#Extracts Kitsune features from given pcap file one packet at a time using "get_next_vector()"
# If wireshark is installed (tshark) it is used to parse (it's faster), otherwise, scapy is used (much slower).
# If wireshark is used then a tsv file (parsed version of the pcap) will be made -which you can use as your input next time
class FE:
    def __init__(self,file_path, parse_type, limit=None, nstat=None, dummy_nstat=None, force_tsv=True, log_file=None):
        self.path = file_path
        self.limit = limit
        self.parse_type = None #unknown
        self.curPacketIndx = 0
        self.input_stream=None
        self.parse_type=parse_type
        self.force_tsv=force_tsv


        if self.parse_type=="tsv":
            # prepare tsv file
            if self.path.endswith("pcap"):
                if not os.path.isfile(self.path+".tsv") or self.force_tsv:
                    self.pcap2tsv_with_tshark()
                self.path+=".tsv"
            self.input_stream=csv.reader(open(self.path, "r"),delimiter="\t")
            #skip header
            next(self.input_stream)

        else:
            ### Prep pcap ##
            print("Reading PCAP file via Scapy...")

            # self.scapyin = rdpcap(self.path, count=self.max_pkt)
            self.input_stream = PcapReader(self.path)

        ### Prep Feature extractor (AfterImage) ###
        self.maxHost = 100000000000
        self.maxSess = 100000000000

        if nstat is not None:
            self.nstat=nstat
            if log_file:
                self.nstat.set_netstat_log_path(log_file)
        else:
            self.nstat = netStat(np.nan, self.maxHost, self.maxSess, log_path=log_file)

        # if dummy_nstat is not None:
        #     self.dummy_nstat=dummy_nstat
        # else:
        #     self.dummy_nstat=ns.netStat(np.nan, self.maxHost, self.maxSess)

    def pcap2tsv_with_tshark(self):
        print('Parsing with tshark...')
        fields = "-e frame.time_epoch -e frame.len -e eth.src -e eth.dst -e ip.src -e ip.dst -e tcp.srcport -e tcp.dstport -e udp.srcport -e udp.dstport -e icmp.type -e icmp.code -e arp.opcode -e arp.src.hw_mac -e arp.src.proto_ipv4 -e arp.dst.hw_mac -e arp.dst.proto_ipv4 -e ipv6.src -e ipv6.dst -e ip.proto -e llc.type"
        if self.limit is not None:
            options=" -c "+ self.limit
        else:
            options=""
        cmd =  'tshark -r '+ self.path +' -T fields '+ fields + options +' -E header=y -E occurrence=f > '+self.path+".tsv"
        subprocess.call(cmd,shell=True)
        print("tshark parsing complete. File saved as: "+self.path +".tsv")


    def get_next_vector(self):
        if self.parse_type=="tsv":
            return self.get_next_vector_tsv()
        elif self.parse_type=="scapy":
            return self.get_next_vector_scapy()

    def get_nstat(self):
        return self.nstat #, self.dummy_nstat

    def get_next_vector_tsv(self):
        row = next(self.input_stream)
        IPtype = np.nan
        packet=row
        if row[-1] == "":
            self.curPacketIndx = self.curPacketIndx + 1
            return []
        timestamp = row[0]
        framelen = row[1]
        srcIP = ''
        dstIP = ''
        if row[4] != '':  # IPv4
            srcIP = row[4]
            dstIP = row[5]
            IPtype = 0
        elif row[17] != '':  # ipv6
            srcIP = row[17]
            dstIP = row[18]
            IPtype = 1
        srcproto = row[6] + row[
            8]  # UDP or TCP port: the concatenation of the two port strings will will results in an OR "[tcp|udp]"
        dstproto = row[7] + row[9]  # UDP or TCP port
        srcMAC = row[2]
        dstMAC = row[3]
        if srcproto == '':  # it's a L2/L1 level protocol
            if row[12] != '':  # is ARP
                srcproto = 'arp'
                dstproto = 'arp'
                srcIP = row[14]  # src IP (ARP)
                dstIP = row[16]  # dst IP (ARP)
                IPtype = 0
            elif row[10] != '':  # is ICMP
                srcproto = 'icmp'
                dstproto = 'icmp'
                IPtype = 0
            elif srcIP + srcproto + dstIP + dstproto == '':  # some other protocol
                srcIP = row[2]  # src MAC
                dstIP = row[3]  # dst MAC

        return  [IPtype, srcMAC, dstMAC, srcIP, srcproto, dstIP, dstproto, float(timestamp), int(framelen)]


    def get_next_vector_scapy(self):
        pkt_tuple = self.input_stream.read_packet()
        packet, pkt_metadata=pkt_tuple[0],pkt_tuple[1]

        # only process IP packets,
        if not (packet.haslayer(IP) or packet.haslayer(IPv6) or packet.haslayer(ARP)):
            return [], packet
        # packet.show2()
        # print(dir(packet))

        timestamp = packet.time
        framelen = len(packet)
        if packet.haslayer(IP):  # IPv4
            srcIP = packet[IP].src
            dstIP = packet[IP].dst
            IPtype = 0
        elif packet.haslayer(IPv6):  # ipv6
            srcIP = packet[IPv6].src
            dstIP = packet[IPv6].dst
            IPtype = 1
        else:
            srcIP = ''
            dstIP = ''

        if packet.haslayer(TCP):
            srcproto = str(packet[TCP].sport)
            dstproto = str(packet[TCP].dport)
        elif packet.haslayer(UDP):
            srcproto = str(packet[UDP].sport)
            dstproto = str(packet[UDP].dport)
        else:
            srcproto = ''
            dstproto = ''

        if packet.haslayer(ARP):
            srcMAC=packet[ARP].hwsrc
            dstMAC=packet[ARP].hwdst
        else:
            srcMAC = packet.src
            dstMAC = packet.dst

        if srcproto == '':  # it's a L2/L1 level protocol
            if packet.haslayer(ARP):  # is ARP
                srcproto = 'arp'
                dstproto = 'arp'
                srcIP = packet[ARP].psrc  # src IP (ARP)
                dstIP = packet[ARP].pdst  # dst IP (ARP)
                IPtype = 0
            elif packet.haslayer(ICMP):  # is ICMP
                srcproto = 'icmp'
                dstproto = 'icmp'
                IPtype = 0
            elif srcIP + srcproto + dstIP + dstproto == '':  # some other protocol
                srcIP = packet.src  # src MAC
                dstIP = packet.dst  # dst MAC

        return [IPtype, srcMAC, dstMAC, srcIP, srcproto, dstIP, dstproto, float(timestamp), int(framelen)], packet

    # def save_nstat_state(self):
    #     f=open('tmp_nstat.txt', 'wb')
    #     pickle.dump( obj=self.nstat,file=f)
    #
    # def roll_back(self):
    #     """Roll back dummy to nstat"""
    #
    #     self.dummy_nstat = pickle.load(open('tmp_nstat.txt', 'rb'))


    def get_num_features(self):
        return len(self.nstat.getNetStatHeaders())



import math
import numpy as np
from pprint import pformat

class incStat:
    def __init__(self, Lambda, ID, init_time=0, isTypeDiff=False):  # timestamp is creation time
        self.ID = ID
        self.CF1 = 0  # linear sum
        self.CF2 = 0  # sum of squares
        self.w = 1e-20  # weight
        self.isTypeDiff = isTypeDiff
        self.Lambda = Lambda  # Decay Factor
        self.lastTimestamp = init_time
        self.cur_mean = np.nan
        self.cur_var = np.nan
        self.cur_std = np.nan
        self.covs = [] # a list of incStat_covs (references) with relate to this incStat

    def __repr__(self):
        return pformat(vars(self))

    def insert(self, v, t=0):  # v is a scalar, t is v's arrival the timestamp
        if self.isTypeDiff:
            dif = t - self.lastTimestamp
            if dif > 0:
                v = dif
            else:
                v = 0
        self.processDecay(t)

        # update with v
        self.CF1 += v
        self.CF2 += math.pow(v, 2)
        self.w += 1
        self.cur_mean = np.nan  # force recalculation if called
        self.cur_var = np.nan
        self.cur_std = np.nan

        # update covs (if any)
        for cov in self.covs:
            cov.update_cov(self.ID, v, t)

    def processDecay(self, timestamp):
        factor=1
        # check for decay
        timeDiff = timestamp - self.lastTimestamp
        if timeDiff > 0:
            factor = math.pow(2, (-self.Lambda * timeDiff))
            self.CF1 = self.CF1 * factor
            self.CF2 = self.CF2 * factor
            self.w = self.w * factor
            self.lastTimestamp = timestamp
        return factor

    def weight(self):
        return self.w

    def mean(self):
        if math.isnan(self.cur_mean):  # calculate it only once when necessary
            self.cur_mean = self.CF1 / self.w
        return self.cur_mean

    def var(self):
        if math.isnan(self.cur_var):  # calculate it only once when necessary
            self.cur_var = abs(self.CF2 / self.w - math.pow(self.mean(), 2))
        return self.cur_var

    def std(self):
        if math.isnan(self.cur_std):  # calculate it only once when necessary
            self.cur_std = math.sqrt(self.var())
        return self.cur_std

    def cov(self,ID2):
        for cov in self.covs:
            if cov.incStats[0].ID == ID2 or cov.incStats[1].ID == ID2:
                return cov.cov()
        return [np.nan]

    def pcc(self,ID2):
        for cov in self.covs:
            if cov.incStats[0].ID == ID2 or cov.incStats[1].ID == ID2:
                return cov.pcc()
        return [np.nan]

    def cov_pcc(self,ID2):
        for cov in self.covs:
            if cov.incStats[0].ID == ID2 or cov.incStats[1].ID == ID2:
                return cov.get_stats1()
        return [np.nan]*2

    def radius(self, other_incStats):  # the radius of a set of incStats
        A = self.var()
        for incS in other_incStats:
            A += incS.var()
        return math.sqrt(A)

    def magnitude(self, other_incStats):  # the magnitude of a set of incStats
        A = math.pow(self.mean(), 2)
        for incS in other_incStats:
            A += math.pow(incS.mean(), 2)
        return math.sqrt(A)

    #calculates and pulls all stats on this stream
    def allstats_1D(self):
        self.cur_mean = self.CF1 / self.w
        self.cur_var = abs(self.CF2 / self.w - math.pow(self.cur_mean, 2))
        return [self.w, self.cur_mean, self.cur_var]

    #calculates and pulls all stats on this stream, and stats shared with the indicated stream
    def allstats_2D(self, ID2):
        stats1D = self.allstats_1D()
        # Find cov component
        stats2D = [np.nan] * 4
        for cov in self.covs:
            if cov.incStats[0].ID == ID2 or cov.incStats[1].ID == ID2:
                stats2D = cov.get_stats2()
                break
        return stats1D + stats2D

    def getHeaders_1D(self, suffix=True):
        if self.ID is None:
            s0=""
        else:
            s0 = "_0"
        if suffix:
            s0 = "_"+self.ID
        headers = ["weight"+s0, "mean"+s0, "std"+s0]
        return headers

    def getHeaders_2D(self, ID2, suffix=True):
        hdrs1D = self.getHeaders_1D(suffix)
        if self.ID is None:
            s0=""
            s1=""
        else:
            s0 = "_0"
            s1 = "_1"
        if suffix:
            s0 = "_"+self.ID
            s1 = "_" + ID2
        hdrs2D = ["radius_" + s0 + "_" + s1, "magnitude_" + s0 + "_" + s1, "covariance_" + s0 + "_" + s1,
                   "pcc_" + s0 + "_" + s1]
        return hdrs1D+hdrs2D


#like incStat, but maintains stats between two streams
class incStat_cov:
    def __init__(self, incS1, incS2, init_time = 0):
        # store references tot he streams' incStats
        self.incStats = [incS1,incS2]
        self.lastRes = [0,0]
        # init extrapolators
        #self.EXs = [extrapolator(),extrapolator()]

        # init sum product residuals
        self.CF3 = 0 # sum of residule products (A-uA)(B-uB)
        self.w3 = 1e-20
        self.lastTimestamp_cf3 = init_time


    def __repr__(self):
        return pformat(self.get_stats1())

    #other_incS_decay is the decay factor of the other incstat
    # ID: the stream ID which produced (v,t)
    def update_cov(self, ID, v, t):  # it is assumes that incStat "ID" has ALREADY been updated with (t,v) [this si performed automatically in method incStat.insert()]
        # find incStat
        if ID == self.incStats[0].ID:
            inc = 0
        elif ID == self.incStats[1].ID:
            inc = 1
        else:
            print("update_cov ID error")
            return ## error

        # Decay other incStat
        self.incStats[not(inc)].processDecay(t)

        # Decay residules
        self.processDecay(t,inc)

        # Update extrapolator for current stream
        #self.EXs[inc].insert(t,v)

        # Extrapolate other stream
        #v_other = self.EXs[not(inc)].predict(t)

        # Compute and update residule
        res = (v - self.incStats[inc].mean())
        resid = (v - self.incStats[inc].mean()) * self.lastRes[not(inc)]
        self.CF3 += resid
        self.w3 += 1
        self.lastRes[inc] = res

    def processDecay(self,t,micro_inc_indx):
        factor = 1
        # check for decay cf3
        timeDiffs_cf3 = t - self.lastTimestamp_cf3
        if timeDiffs_cf3 > 0:
            factor = math.pow(2, (-(self.incStats[micro_inc_indx].Lambda) * timeDiffs_cf3))
            self.CF3 *= factor
            self.w3 *= factor
            self.lastTimestamp_cf3 = t
            self.lastRes[micro_inc_indx] *= factor
        return factor

    #todo: add W3 for cf3

    #covariance approximation
    def cov(self):
        return self.CF3 / self.w3

    # Pearson corl. coef
    def pcc(self):
        ss = self.incStats[0].std() * self.incStats[1].std()
        if ss != 0:
            return self.cov() / ss
        else:
            return 0

    # calculates and pulls all correlative stats
    def get_stats1(self):
        return [self.cov(), self.pcc()]

    # calculates and pulls all correlative stats AND 2D stats from both streams (incStat)
    def get_stats2(self):
        return [self.incStats[0].radius([self.incStats[1]]),self.incStats[0].magnitude([self.incStats[1]]),self.cov(), self.pcc()]

    # calculates and pulls all correlative stats AND 2D stats AND the regular stats from both streams (incStat)
    def get_stats3(self):
        return [self.incStats[0].w,self.incStats[0].mean(),self.incStats[0].std(),self.incStats[1].w,self.incStats[1].mean(),self.incStats[1].std(),self.cov(), self.pcc()]

    # calculates and pulls all correlative stats AND the regular stats from both incStats AND 2D stats
    def get_stats4(self):
        return [self.incStats[0].w,self.incStats[0].mean(),self.incStats[0].std(),self.incStats[1].w,self.incStats[1].mean(),self.incStats[1].std(), self.incStats[0].radius([self.incStats[1]]),self.incStats[0].magnitude([self.incStats[1]]),self.cov(), self.pcc()]

    def getHeaders(self,ver,suffix=True): #ver = {1,2,3,4}
        headers = []
        s0 = "0"
        s1 = "1"
        if suffix:
            s0 = self.incStats[0].ID
            s1 = self.incStats[1].ID

        if ver == 1:
            headers = ["covariance_"+s0+"_"+s1, "pcc_"+s0+"_"+s1]
        if ver == 2:
            headers = ["radius_"+s0+"_"+s1, "magnitude_"+s0+"_"+s1, "covariance_"+s0+"_"+s1, "pcc_"+s0+"_"+s1]
        if ver == 3:
            headers = ["weight_"+s0, "mean_"+s0, "std_"+s0,"weight_"+s1, "mean_"+s1, "std_"+s1, "covariance_"+s0+"_"+s1, "pcc_"+s0+"_"+s1]
        if ver == 4:
            headers = ["weight_" + s0, "mean_" + s0, "std_" + s0, "covariance_" + s0 + "_" + s1, "pcc_" + s0 + "_" + s1]
        if ver == 5:
            headers = ["weight_"+s0, "mean_"+s0, "std_"+s0,"weight_"+s1, "mean_"+s1, "std_"+s1, "radius_"+s0+"_"+s1, "magnitude_"+s0+"_"+s1, "covariance_"+s0+"_"+s1, "pcc_"+s0+"_"+s1]
        return headers


class incStatDB:
    # default_lambda: use this as the lambda for all streams. If not specified, then you must supply a Lambda with every query.
    def __init__(self,limit=np.Inf,default_lambda=np.nan):
        self.HT = dict()
        self.limit = limit
        self.df_lambda = default_lambda

    def get_lambda(self,Lambda):
        if not np.isnan(self.df_lambda):
            Lambda = self.df_lambda
        return Lambda

    def __repr__(self):
        return pformat(self.HT,indent=2)


    # Registers a new stream. init_time: init lastTimestamp of the incStat
    def register(self,ID,Lambda=1,init_time=0,isTypeDiff=False):
        #Default Lambda?
        Lambda = self.get_lambda(Lambda)

        #Retrieve incStat
        key = ID+"_"+str(Lambda)
        incS = self.HT.get(key)
        if incS is None: #does not already exist
            if len(self.HT) + 1 > self.limit:
                raise LookupError(
                    'Adding Entry:\n' + key + '\nwould exceed incStatHT 1D limit of ' + str(
                        self.limit) + '.\nObservation Rejected.')
            incS = incStat(Lambda, ID, init_time, isTypeDiff)
            self.HT[key] = incS #add new entry
        return incS

    # Registers covariance tracking for two streams, registers missing streams
    def register_cov(self,ID1,ID2,Lambda=1,init_time=0,isTypeDiff=False):
        #Default Lambda?
        Lambda = self.get_lambda(Lambda)

        # Lookup both streams
        incS1 = self.register(ID1,Lambda,init_time,isTypeDiff)
        incS2 = self.register(ID2,Lambda,init_time,isTypeDiff)

        #check for pre-exiting link
        for cov in incS1.covs:
            if cov.incStats[0].ID == ID2 or cov.incStats[1].ID == ID2:
                return cov #there is a pre-exiting link

        # Link incStats
        inc_cov = incStat_cov(incS1,incS2,init_time)
        incS1.covs.append(inc_cov)
        incS2.covs.append(inc_cov)
        return inc_cov

    # updates/registers stream
    def update(self,ID,t,v,Lambda=1,isTypeDiff=False):
        incS = self.register(ID,Lambda,t,isTypeDiff)
        incS.insert(v,t)
        return incS

    # Pulls current stats from the given ID
    def get_1D_Stats(self,ID,Lambda=1): #weight, mean, std
        #Default Lambda?
        Lambda = self.get_lambda(Lambda)

        #Get incStat
        incS = self.HT.get(ID+"_"+str(Lambda))
        if incS is None:  # does not already exist
            return [np.na]*3
        else:
            return incS.allstats_1D()

    # Pulls current correlational stats from the given IDs
    def get_2D_Stats(self, ID1, ID2, Lambda=1): #cov, pcc
        # Default Lambda?
        Lambda = self.get_lambda(Lambda)

        # Get incStat
        incS1 = self.HT.get(ID1 + "_" + str(Lambda))
        if incS1 is None:  # does not exist
            return [np.na]*2

        # find relevant cov entry
        return incS1.cov_pcc(ID2)

    # Pulls all correlational stats registered with the given ID
    # returns tuple [0]: stats-covs&pccs, [2]: IDs
    def get_all_2D_Stats(self, ID, Lambda=1):  # cov, pcc
        # Default Lambda?
        Lambda = self.get_lambda(Lambda)

        # Get incStat
        incS1 = self.HT.get(ID + "_" + str(Lambda))
        if incS1 is None:  # does not exist
            return ([],[])

        # find relevant cov entry
        stats = []
        IDs = []
        for cov in incS1.covs:
            stats.append(cov.get_stats1())
            IDs.append([cov.incStats[0].ID,cov.incStats[1].ID])
        return stats,IDs

    # Pulls current multidimensional stats from the given IDs
    def get_nD_Stats(self,IDs,Lambda=1): #radius, magnitude (IDs is a list)
        # Default Lambda?
        Lambda = self.get_lambda(Lambda)

        # Get incStats
        incStats = []
        for ID in IDs:
            incS = self.HT.get(ID + "_" + str(Lambda))
            if incS is not None:  #exists
                incStats.append(incS)

        # Compute stats
        rad = 0 #radius
        mag = 0 #magnitude
        for incS in incStats:
            rad += incS.var()
            mag += incS.mean()**2

        return [np.sqrt(rad),np.sqrt(mag)]

    # Updates and then pulls current 1D stats from the given ID. Automatically registers previously unknown stream IDs
    def update_get_1D_Stats(self, ID,t,v,Lambda=1,isTypeDiff=False):  # weight, mean, std
        incS = self.update(ID,t,v,Lambda,isTypeDiff)
        return incS.allstats_1D()


    # Updates and then pulls current correlative stats between the given IDs. Automatically registers previously unknown stream IDs, and cov tracking
    #Note: AfterImage does not currently support Diff Type streams for correlational statistics.
    def update_get_2D_Stats(self, ID1,ID2,t1,v1,Lambda=1,level=1):  #level=  1:cov,pcc  2:radius,magnitude,cov,pcc
        #retrieve/add cov tracker
        inc_cov = self.register_cov(ID1, ID2, Lambda,  t1)
        # Update cov tracker
        inc_cov.update_cov(ID1,v1,t1)
        if level == 1:
            return inc_cov.get_stats1()
        else:
            return inc_cov.get_stats2()

    def get_2D_stats(self, ID1, ID2, Lambda, t1, level=1):
        inc_cov = self.register_cov(ID1, ID2, Lambda,  t1)
        if level == 1:
            return inc_cov.get_stats1()
        else:
            return inc_cov.get_stats2()

    # Updates and then pulls current 1D and 2D stats from the given IDs. Automatically registers previously unknown stream IDs
    def update_get_1D2D_Stats(self, ID1,ID2,t1,v1,Lambda=1):  # weight, mean, std
        return self.update_get_1D_Stats(ID1,t1,v1,Lambda) + self.update_get_2D_Stats(ID1,ID2,t1,v1,Lambda,level=2)

    def getHeaders_1D(self,Lambda=1,ID=None):
        # Default Lambda?
        Lambda = self.get_lambda(Lambda)
        hdrs = incStat(Lambda,ID).getHeaders_1D(suffix=False)
        return [str(Lambda)+"_"+s for s in hdrs]

    def getHeaders_2D(self,Lambda=1,IDs=None, ver=1): #IDs is a 2-element list or tuple
        # Default Lambda?
        Lambda = self.get_lambda(Lambda)
        if IDs is None:
            IDs = [0,1]
        hdrs = incStat_cov(incStat(Lambda,IDs[0]),incStat(Lambda,IDs[0]),Lambda).getHeaders(ver,suffix=False)
        return [str(Lambda)+"_"+s for s in hdrs]

    def getHeaders_1D2D(self,Lambda=1,IDs=None, ver=1):
        # Default Lambda?
        Lambda = self.get_lambda(Lambda)
        if IDs is None:
            IDs = [0,1]
        hdrs1D = self.getHeaders_1D(Lambda,IDs[0])
        hdrs2D = self.getHeaders_2D(Lambda,IDs, ver)
        return hdrs1D + hdrs2D

    def getHeaders_nD(self,Lambda=1,IDs=[]): #IDs is a n-element list or tuple
        # Default Lambda?
        ID = ":"
        for s in IDs:
            ID += "_"+s
        Lambda = self.get_lambda(Lambda)
        hdrs = ["radius"+ID, "magnitude"+ID]
        return [str(Lambda)+"_"+s for s in hdrs]


    #cleans out records that have a weight less than the cutoff.
    #returns number or removed records.
    def cleanOutOldRecords(self,cutoffWeight,curTime):
        n = 0
        dump = sorted(self.HT.items(), key=lambda tup: tup[1][0].getMaxW(curTime))
        for entry in dump:
            entry[1][0].processDecay(curTime)
            W = entry[1][0].w
            if W <= cutoffWeight:
                key = entry[0]
                del entry[1][0]
                del self.HT[key]
                n=n+1
            elif W > cutoffWeight:
                break
        return n



import enum
from itertools import zip_longest
from typing import IO, Any, Dict, Iterator, List, Optional, Tuple, Union

import pandas as pd
import torch
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import FileLister, FileOpener, IterDataPipe

ParquetRow = Dict[str, Any]
ParquetParserReturnType = Union[ParquetRow, Tuple[str, Any]]


class FileFormat(enum.Enum):
    csv: str = "csv"
    parquet: str = "parquet"


@functional_datapipe("parse_parquet")
class ParquetParserIterDataPipe(IterDataPipe[ParquetParserReturnType]):

    def __init__(self,
                 source_datapipe: IterDataPipe[Tuple[str, IO]],
                 *,
                 return_path: bool = False,
                 engine: str = "auto",
                 columns: Optional[List[str]] = None) -> None:
        super().__init__()
        self.source_datapipe = source_datapipe
        self.return_path = return_path
        self.engine = engine
        self.columns = columns

    def __iter__(self) -> Iterator[ParquetParserReturnType]:
        for path, io_file in self.source_datapipe:
            parquet_df = pd.read_parquet(io_file,
                                         engine=self.engine,
                                         columns=self.columns)

            it = (o[1].to_dict() for o in parquet_df.iterrows())
            if self.return_path:
                yield from zip_longest([path], it)
            else:
                yield from it


def build_input_data_pipe(
        root: str,
        batch_size: int,
        shuffle: bool = True,
        file_format: FileFormat = FileFormat.csv) -> IterDataPipe[torch.Tensor]:

    file_filter = lambda fname: fname.endswith(file_format.value)
    dp = FileLister(root=root, recursive=True).filter(file_filter)
    dp = FileOpener(dp, mode="rb")

    if file_format == FileFormat.parquet:
        dp = dp.parse_parquet(return_path=False, columns=["features"])
        dp = dp.map(_to_dense_vector)
    else:
        dp = dp.parse_csv(return_path=False)
        dp = dp.map(lambda x: torch.as_tensor([float(o) for o in x]))

    if shuffle:
        dp = dp.shuffle()

    dp = dp.batch(batch_size=batch_size, drop_last=False)
    return dp.collate(torch.stack)


def _to_dense_vector(row: ParquetRow) -> torch.Tensor:
    sample_features = row["features"]
    if sample_features["indices"] is not None:
        features = torch.zeros((int(sample_features["size"]),)).float()
        indices = torch.as_tensor(sample_features["indices"].copy()).long()
        features[indices] = torch.as_tensor(
            sample_features["values"].copy()).float()
    else:
        features = torch.as_tensor(sample_features["values"].copy()).float()
    return features


if __name__ == "__main__":
    # parquet_dp = FileLister(root="data/parquet")
    dp = build_input_data_pipe(root="data/kitnet-test",
                               batch_size=16,
                               file_format=FileFormat.csv)
    for features in dp:
        print(features.size())
        break



from textwrap import fill
import datetime
import matplotlib.ticker as ticker
import matplotlib.dates as mdate
import sklearn.metrics as metrics
from itertools import product
from tqdm import tqdm
from matplotlib import cm
from matplotlib import pyplot as plt
from scipy.stats import norm
import numpy as np
import pickle
from .KitNET.KitNET import KitNET
import matplotlib
import socket
from .feature_squeeze import squeeze_features
import multiprocessing as mp
import csv
matplotlib.use('Agg')
np.set_printoptions(threshold=np.inf)
# matplotlib.rcParams['timezone']="Pacific/Auckland"


def train_normal(params):
    """
    trains kitsune on normal traffic

    Args:
        params (dict): set of training parameters.

    Returns:
        type: None, model is saved in model_path


    """
    # Build Kitsune
    K = KitNET(100, params["maxAE"], params["FMgrace"],
               params["ADgrace"], 0.1, 0.75, normalize=params["normalize"])

    input_file = open(params["path"], "r")
    input_file.readline()
    count = 0
    tbar = tqdm()
    rmse = []
    while True:
        feature_vector = input_file.readline()
        # check EOF via empty string
        if not feature_vector:
            break
        fv = feature_vector.rstrip().split(",")
        fv = fv[:100]
        fv = np.array(fv, dtype="float")
        K.process(fv)
        count += 1
        tbar.update(1)
        if count > params["FMgrace"] + params["ADgrace"]:
            break
    tbar.close()
    # save
    with open(params["model_path"], "wb") as of:
        pickle.dump(K, of)


def eval_kitsune(path, model_path, threshold=None, ignore_index=-1, out_image=None, meta_file=None, record_scores=False, y_true=None, record_prediction=False, load_prediction=False, plot_with_time=False):
    """
    evaluates trained kitsune model on some traffic.

    Args:
        path (string): path to traffic feature file.
        model_path (string): path to trained kitsune model.
        threshold (float): anomaly threshold value, if None it calculates the threshold value as 3 std away from mean. Defaults to None.
        ignore_index (int): number of features to ignore at the start. Defaults to -1.
        out_image (string): path to output anomaly score image. Defaults to None.
        meta_file (string): path to metadata file, used to calculate evasion metrics. Defaults to None.
        record_scores (boolean): whether to record anomaly scores in a seperate csv file. Defaults to False.

    Returns:
        if has_meta: return number of positive samples and positive samples that are not craft packets.
        else: return number of positive samples

    """
    # the pcap, pcapng, or tsv file to process.
    print("evaluting", path)
    print("meta", meta_file)
    print("kitsune model path ", model_path)
    # if(threshold is not None):
    #     t = open(threshold, "r")
    #     threshold = t.readline() 
    # else:
    #     t = threshold 
    t = threshold 

    roc_auc = 1
    label_map = []

    with open(model_path, "rb") as m:
        kitsune = pickle.load(m)

    if out_image == None:

        out_image = path[:-4] + "_kitsune_rmse.png"

    if meta_file is not None:
        meta = open(meta_file, "r")
        meta.readline()
        meta_row = meta.readline()
        has_meta = True
        pos_craft = 0
        pos_mal = 0
        pos_ignore = 0
    else:
        has_meta = False
        pos = 0

    labels = []
    times = []
    colours = []
    tbar = tqdm()
    if load_prediction:
        rmse_array = np.genfromtxt(
            path[:-4] + "_kitsune_score.csv", delimiter=",")
    else:
        counter = 0
        input_file = open(path, "r")
        input_file.readline()
        rmse_array = []

        if not has_meta:
            colours = None

        feature_vector = input_file.readline()
        while feature_vector is not '':

            if counter < ignore_index:
                feature_vector = input_file.readline()

                if meta_file is not None:
                    meta_row = meta.readline()

                counter += 1
                continue

            fv = feature_vector.rstrip().split(",")

            if len(fv) == 102:
                label = fv[-2]
                if label.isdigit():
                    try:
                        label = socket.getservbyport(int(label))
                    except OSError:
                        label = "udp"
                if label not in label_map:
                    label_map.append(label)

                labels.append(label_map.index(label))
                times.append(mdate.epoch2num(float(fv[-1])) + 1)
                fv = fv[:100]

            fv = np.array(fv, dtype="float")

            try:
                if kitsune.input_precision is not None:
                    fv = squeeze_features(fv, kitsune.input_precision)
            except AttributeError as e:
                pass

            rmse = kitsune.process(fv)

            if rmse == 0:

                rmse_array.append(1e-2)
            else:

                rmse_array.append(rmse)
            counter += 1
            tbar.update(1)

            feature_vector = input_file.readline()
            # set colours
            if has_meta:
                comment = meta_row.rstrip().split(",")[-1]
                if comment == "craft":
                    colours.append([67 / 255., 67 / 255., 67 / 255., 0.8])

                elif comment == "malicious":
                    colours.append([1, 0, 0, 1])
                else:
                    colours.append([204 / 255., 243 / 255., 1, 0.5])

            if threshold is not None and rmse > float(threshold):
                if has_meta:
                    comment = meta_row.rstrip().split(",")[-1]
                    if comment == "craft":
                        pos_craft += 1
                    elif comment == "malicious":
                        pos_mal += 1
                    elif comment == "attacker_low":
                        pos_ignore += 1
                    else:
                        print(meta_row)
                        print(rmse)
                        raise Exception
                else:
                    pos += 1

            if has_meta:
                meta_row = meta.readline()

    # if no threshold, calculate threshold
    if threshold == None:
        # threshold is min(mean+3std, max)
        benignSample = np.log(rmse_array)
        mean = np.mean(benignSample)
        std = np.std(benignSample)
        threshold_std = np.exp(mean + 3 * std)
        threshold_max = max(rmse_array)
        threshold = min(threshold_max, threshold_std)
        pos = (rmse_array > threshold).sum()

    # record prediction scores/rmse
    if record_scores:
        score_path = "_kitsune_score.csv"
        threshold_path = "_kitsune_threshold.csv"
        # print("max_rmse",np.max(rmse_array))
        np.savetxt(score_path, rmse_array, delimiter=",")
        np.savetxt(threshold_path, [float(threshold)], delimiter=",")
        print("score saved to", score_path)

    # record prediction labels
    if record_prediction:
        pred_path = "_kitsune_prediction.csv"
        np.savetxt(pred_path, rmse_array > threshold, delimiter=",")
        print("kitsune prediction saved to", pred_path)

    if y_true is None:

        fpr, tpr, roc_t = metrics.roc_curve(
            [0 for i in range(len(rmse_array))], rmse_array, drop_intermediate=False)
    else:
        fpr, tpr, roc_t = metrics.roc_curve(
            y_true, rmse_array, drop_intermediate=True)
        roc_auc = metrics.auc(fpr, tpr)
    print("total packets:", len(rmse_array))

    # if out_image is not None:
    #     cmap = plt.get_cmap('Set3')
    #     num_packets = len(rmse_array)
    #     f, (ax1, ax2) = plt.subplots(
    #         2, 1, constrained_layout=True, figsize=(10, 10), dpi=200)

    #     if times and plot_with_time:
    #         x_val = times
    #         date_fmt = '%m/%d %H:%M:%S'

    #         date_formatter = mdate.DateFormatter(date_fmt)
    #         ax1.xaxis.set_major_formatter(date_formatter)

    #         # tick every 4 hours
    #         # print("asdfs")
    #         ax1.xaxis.set_major_locator(ticker.MultipleLocator(0.85))

    #         ax1.tick_params(labelrotation=90)
    #         # f.autofmt_xdate()
    #     else:
    #         x_val = range(len(rmse_array))

    #     if labels:
    #         (unique, counts) = np.unique(labels, return_counts=True)
    #         frequencies = np.asarray((unique, counts)).T
    #         for i in frequencies:
    #             label_map[i[0]] = "{} {}".format(label_map[i[0]], i[1])

    #         scatter = ax1.scatter(x_val, rmse_array,
    #                               s=1, c=labels, alpha=0.8, cmap=cmap)
    #         # wrap legends
    #         labels = [fill(l, 20) for l in label_map]

    #         leg = ax1.legend(handles=scatter.legend_elements()[0], labels=labels, bbox_to_anchor=(1.01, 1),
    #                          loc='upper left', borderaxespad=0.)
    #         leg.get_lines()[0].set_alpha(1.)
            
    #     elif has_meta:
    #         ax1.scatter(x_val, rmse_array, s=1, c='#00008B')
    #     else:
    #         ax1.scatter(x_val, rmse_array, s=1, alpha=0.8 , c = '#FF8C00')

    #     # max_rmse=np.max(rmse_array)
    #     # print(max_rmse)

    #     ax1.axhline(y=threshold, color='r', linestyle='-')
    #     ax1.set_yscale("log")
    #     ax1.set_title("Anomaly Scores from Kitsune_{} Execution Phase".format(
    #         model_path.split("/")[-1]))
    #     ax1.set_ylabel("RMSE (log scaled)")
    #     if has_meta:
    #         ax1.set_xlabel(
    #             "packet index \n packets over threshold {}".format(pos_mal + pos_craft))
    #     else:
    #         ax1.set_xlabel(
    #             "packet index \n packets over threshold {}".format(pos))

    #     if y_true is None:
    #         ax2.plot(fpr, roc_t, 'b')
    #         ax2.set_ylabel("threshold")
    #         ax2.set_xlabel("false positive rate")
    #     else:
    #         ax2.plot(fpr, tpr, 'b', label='AUC = %0.2f' % roc_auc)
    #         ax2.set_title('AUC = %0.2f' % roc_auc)
    #         ax2.set_ylabel("true positive rate")
    #         ax2.set_xlabel("false positive rate")
    #     # plt.tight_layout()
    #     f.savefig(out_image)
    #     print("plot path:", out_image)
    #     plt.close()
    # tbar.close()
    # if has_meta:
    #     return pos_mal, pos_craft, pos_ignore
    # else:
    #     if t is None:
    #         return pos, threshold
    #     else:
    #         return pos, roc_auc

    # From here I am adding my own code for plotting the graph
    if out_image is not None:
        cmap = plt.get_cmap('Set3')
        num_packets = len(rmse_array)
        f, (ax1, ax2) = plt.subplots(
            2, 1, constrained_layout=True, figsize=(10, 10), dpi=200)

        if times and plot_with_time:
            x_val = times
            date_fmt = '%m/%d %H:%M:%S'

            date_formatter = mdate.DateFormatter(date_fmt)
            ax1.xaxis.set_major_formatter(date_formatter)

        # tick every 4 hours
        # print("asdfs")
            ax1.xaxis.set_major_locator(ticker.MultipleLocator(0.85))

            ax1.tick_params(labelrotation=90)
        # f.autofmt_xdate()
        else:
            x_val = range(len(rmse_array))

        if labels:
            (unique, counts) = np.unique(labels, return_counts=True)
            frequencies = np.asarray((unique, counts)).T
            for i in frequencies:
                label_map[i[0]] = "{} {}".format(label_map[i[0]], i[1])

            scatter = ax1.scatter(x_val, rmse_array,
                              s=1, c=labels, alpha=1.0, cmap=cmap)  # Set alpha to 1.0
        # wrap legends
            labels = [fill(l, 20) for l in label_map]

            leg = ax1.legend(handles=scatter.legend_elements()[0], labels=labels, bbox_to_anchor=(1.01, 1),
                             loc='upper left', borderaxespad=0.)
            leg.get_lines()[0].set_alpha(1.0)  # Set alpha to 1.0

    elif has_meta:
        ax1.scatter(x_val, rmse_array, s=1, c='#00008B')
    else:
        ax1.scatter(x_val, rmse_array, s=1, alpha=1.0, c='#FF8C00')  # Set alpha to 1.0

    # max_rmse=np.max(rmse_array)
    # print(max_rmse)

    ax1.axhline(y=threshold, color='r', linestyle='-')
    ax1.set_yscale("log")
    ax1.set_title("Anomaly Scores from Kitsune_{} Execution Phase".format(
        path.split("/")[-1]))
    ax1.set_ylabel("RMSE (log scaled)")
    if has_meta:
        ax1.set_xlabel(
            "packet index \n packets over threshold {}".format(pos_mal + pos_craft))
    else:
        ax1.set_xlabel(
            "packet index \n packets over threshold {}".format(pos))

    if y_true is None:
        ax2.plot(fpr, roc_t, 'b')
        ax2.set_ylabel("threshold")
        ax2.set_xlabel("false positive rate")
    else:
        ax2.plot(fpr, tpr, 'b', label='AUC = %0.2f' % roc_auc)
        ax2.set_title('AUC = %0.2f' % roc_auc)
        ax2.set_ylabel("true positive rate")
        ax2.set_xlabel("false positive rate")
    # plt.tight_layout()
    f.savefig(out_image)
    print("plot path:", out_image)
    plt.close()
    tbar.close()
    if has_meta:
        return pos_mal, pos_craft, pos_ignore
    else:
        if t is None:
            return pos, threshold
        else:
            return pos, roc_auc


def evaluation_metrics(scores_path, threshold_path):
    # Calulate the F1 score for Model using ground labels
    # Actual outputs from kitsune_score.csv
    y_output = csv.reader(open(scores_path, 'r'))
    y_output = list(y_output)
    y_output = [float(i[0]) for i in y_output]

    # let ground labels be 0(benign) for all pkts
    y_true = np.ones(len(y_output))
    # print(y_true)

    # Threshold for F1 score
    threshold = csv.reader(open(threshold_path, 'r'))
    threshold = list(threshold)
    threshold = float(threshold[0][0])

    # Predicted labels
    y_pred = np.array([0 if i > threshold else 1 for i in y_output])
    # print(y_pred)
    # 0 -> benign and 1 -> malicious
    tp = fp = tn = fn = 0
    for i in range(len(y_true)):
        if y_true[i] == 1 and y_pred[i] == 1:
            tp += 1
        elif y_true[i] == 0 and y_pred[i] == 1:
            fp += 1
        elif y_true[i] == 0 and y_pred[i] == 0:
            tn += 1
        else:
            fn += 1
    # print(metrics.confusion_matrix(y_true, y_pred).ravel())
    # m = metrics.confusion_matrix(y_true, y_pred)
    # tn = m[0][0]
    # fp = m[0][1]
    # fn = m[1][0]
    # tp = m[1][1]
    # tn, fp, fn, tp = metrics.confusion_matrix(y_true, y_pred).ravel()
    precision = tp/(tp+fp)
    recall = tp/(tp+fn)
    f1 = 2*(precision*recall)/(precision+recall)

    print("True Negatives: ", tn)
    print("False Positives: ", fp)
    print("False Negatives: ", fn)
    print("True Positives: ", tp)

    print("Precision: ", precision)
    print("Recall: ", recall)
    print("F1 score: ", f1)


#Check if cython code has been compiled
import os
import subprocess

# use_extrapolation=False #experimental correlation code
# if use_extrapolation:
#     print("Importing AfterImage Cython Library")
#     if not os.path.isfile("AfterImage.c"): #has not yet been compiled, so try to do so...
#         cmd = "python setup.py build_ext --inplace"
#         subprocess.call(cmd,shell=True)
#Import dependencies
from .netStat import netStat as ns
import csv
import numpy as np
print("Importing Scapy Library")
from scapy.all import *
import os.path
import platform
import subprocess


#Extracts Kitsune features from given pcap file one packet at a time using "get_next_vector()"
# If wireshark is installed (tshark) it is used to parse (it's faster), otherwise, scapy is used (much slower).
# If wireshark is used then a tsv file (parsed version of the pcap) will be made -which you can use as your input next time
class FE:
    def __init__(self,file_path,limit=np.inf):
        self.path = file_path
        self.limit = limit
        self.parse_type = None #unknown
        self.curPacketIndx = 0
        self.tsvin = None #used for parsing TSV file
        self.scapyin = None #used for parsing pcap with scapy

        ### Prep pcap ##
        self.__prep__()

        ### Prep Feature extractor (AfterImage) ###
        maxHost = 100000000000
        maxSess = 100000000000
        self.nstat = ns.netStat(np.nan, maxHost, maxSess)

    def _get_tshark_path(self):
        if platform.system() == 'Windows':
            return 'C:\Program Files\Wireshark\\tshark.exe'
        else:
            system_path = os.environ['PATH']
            for path in system_path.split(os.pathsep):
                filename = os.path.join(path, 'tshark')
                if os.path.isfile(filename):
                    return filename
        return ''

    def __prep__(self):
        ### Find file: ###
        if not os.path.isfile(self.path):  # file does not exist
            print("File: " + self.path + " does not exist")
            raise Exception()

        ### check file type ###
        type = self.path.split('.')[-1]

        self._tshark = self._get_tshark_path()
        ##If file is TSV (pre-parsed by wireshark script)
        if type == "tsv":
            self.parse_type = "tsv"

        ##If file is pcap
        elif type == "pcap" or type == 'pcapng':
            # Try parsing via tshark dll of wireshark (faster)
            if os.path.isfile(self._tshark):
                self.pcap2tsv_with_tshark()  # creates local tsv file
                self.path += ".tsv"
                self.parse_type = "tsv"
            else: # Otherwise, parse with scapy (slower)
                print("tshark not found. Trying scapy...")
                self.parse_type = "scapy"
        else:
            print("File: " + self.path + " is not a tsv or pcap file")
            raise Exception()

        ### open readers ##
        if self.parse_type == "tsv":
            maxInt = sys.maxsize
            decrement = True
            while decrement:
                # decrease the maxInt value by factor 10
                # as long as the OverflowError occurs.
                decrement = False
                try:
                    csv.field_size_limit(maxInt)
                except OverflowError:
                    maxInt = int(maxInt / 10)
                    decrement = True

            print("counting lines in file...")
            num_lines = sum(1 for line in open(self.path))
            print("There are " + str(num_lines) + " Packets.")
            self.limit = min(self.limit, num_lines-1)
            self.tsvinf = open(self.path, 'rt', encoding="utf8")
            self.tsvin = csv.reader(self.tsvinf, delimiter='\t')
            row = self.tsvin.__next__() #move iterator past header

        else: # scapy
            print("Reading PCAP file via Scapy...")
            self.scapyin = rdpcap(self.path)
            self.limit = len(self.scapyin)
            print("Loaded " + str(len(self.scapyin)) + " Packets.")

    def get_next_vector(self):
        if self.curPacketIndx == self.limit:
            if self.parse_type == 'tsv':
                self.tsvinf.close()
            return []

        ### Parse next packet ###
        if self.parse_type == "tsv":
            row = self.tsvin.__next__()
            IPtype = np.nan
            timestamp = row[0]
            framelen = row[1]
            srcIP = ''
            dstIP = ''
            if row[4] != '':  # IPv4
                srcIP = row[4]
                dstIP = row[5]
                IPtype = 0
            elif row[17] != '':  # ipv6
                srcIP = row[17]
                dstIP = row[18]
                IPtype = 1
            srcproto = row[6] + row[
                8]  # UDP or TCP port: the concatenation of the two port strings will will results in an OR "[tcp|udp]"
            dstproto = row[7] + row[9]  # UDP or TCP port
            srcMAC = row[2]
            dstMAC = row[3]
            if srcproto == '':  # it's a L2/L1 level protocol
                if row[12] != '':  # is ARP
                    srcproto = 'arp'
                    dstproto = 'arp'
                    srcIP = row[14]  # src IP (ARP)
                    dstIP = row[16]  # dst IP (ARP)
                    IPtype = 0
                elif row[10] != '':  # is ICMP
                    srcproto = 'icmp'
                    dstproto = 'icmp'
                    IPtype = 0
                elif srcIP + srcproto + dstIP + dstproto == '':  # some other protocol
                    srcIP = row[2]  # src MAC
                    dstIP = row[3]  # dst MAC

        elif self.parse_type == "scapy":
            packet = self.scapyin[self.curPacketIndx]
            IPtype = np.nan
            timestamp = packet.time
            framelen = len(packet)
            if packet.haslayer(IP):  # IPv4
                srcIP = packet[IP].src
                dstIP = packet[IP].dst
                IPtype = 0
            elif packet.haslayer(IPv6):  # ipv6
                srcIP = packet[IPv6].src
                dstIP = packet[IPv6].dst
                IPtype = 1
            else:
                srcIP = ''
                dstIP = ''

            if packet.haslayer(TCP):
                srcproto = str(packet[TCP].sport)
                dstproto = str(packet[TCP].dport)
            elif packet.haslayer(UDP):
                srcproto = str(packet[UDP].sport)
                dstproto = str(packet[UDP].dport)
            else:
                srcproto = ''
                dstproto = ''

            srcMAC = packet.src
            dstMAC = packet.dst
            if srcproto == '':  # it's a L2/L1 level protocol
                if packet.haslayer(ARP):  # is ARP
                    srcproto = 'arp'
                    dstproto = 'arp'
                    srcIP = packet[ARP].psrc  # src IP (ARP)
                    dstIP = packet[ARP].pdst  # dst IP (ARP)
                    IPtype = 0
                elif packet.haslayer(ICMP):  # is ICMP
                    srcproto = 'icmp'
                    dstproto = 'icmp'
                    IPtype = 0
                elif srcIP + srcproto + dstIP + dstproto == '':  # some other protocol
                    srcIP = packet.src  # src MAC
                    dstIP = packet.dst  # dst MAC
        else:
            return []

        self.curPacketIndx = self.curPacketIndx + 1


        ### Extract Features
        try:
            return self.nstat.updateGetStats(IPtype, srcMAC, dstMAC, srcIP, srcproto, dstIP, dstproto,
                                                 int(framelen),
                                                 float(timestamp))
        except Exception as e:
            print(e)
            return []


    def pcap2tsv_with_tshark(self):
        print('Parsing with tshark...')
        fields = "-e frame.time_epoch -e frame.len -e eth.src -e eth.dst -e ip.src -e ip.dst -e tcp.srcport -e tcp.dstport -e udp.srcport -e udp.dstport -e icmp.type -e icmp.code -e arp.opcode -e arp.src.hw_mac -e arp.src.proto_ipv4 -e arp.dst.hw_mac -e arp.dst.proto_ipv4 -e ipv6.src -e ipv6.dst"
        cmd =  '"' + self._tshark + '" -r '+ self.path +' -T fields '+ fields +' -E header=y -E occurrence=f > '+self.path+".tsv"
        subprocess.call(cmd,shell=True)
        print("tshark parsing complete. File saved as: "+self.path +".tsv")

    def get_num_features(self):
        return len(self.nstat.getNetStatHeaders())



import math
from itertools import product
from tqdm import tqdm
from matplotlib import cm
from matplotlib import pyplot as plt
from scipy.stats import norm
import numpy as np
import pickle
from .KitNET import *
import sys
sys.path.append("..")


def squeeze_features(fv, precision):
    """rounds features to siginificant figures

    Args:
        fv (array): feature vector.
        precision (int): number of precisions to use.

    Returns:
        array: rounded array of floats.

    """
    fv_positive = np.where(np.isfinite(fv) & (
        fv != 0), np.abs(fv), 10**(precision-1))
    mags = 10 ** (precision - 1 - np.floor(np.log10(fv_positive)))
    return np.round(fv * mags) / mags


def draw_plot(list_to_draw, out_name, threshold=None):
    plt.scatter(np.arange(0, len(list_to_draw), 1),
                list_to_draw,
                s=10,
                label="RMSE")
    if threshold is not None:
        plt.axhline(threshold)
    plt.savefig(out_name)
    plt.clf()


def draw_hist(list_to_draw, out_name):
    plt.hist(list_to_draw)
    plt.yscale('log')
    plt.savefig('fig/{}.png'.format(out_name))
    plt.clf()


def eval_feature_squeeze(path, model_path, out_name, precision, threshold=None):
    """
    evaluates adversarial detection defenses on the trained model.

    Args:
        path (string): path to traffic feature file.
        model_path (string): path to trained kitsune model.

    Returns:
        if has_meta: return number of positive samples and positive samples that are not craft packets.
        else: return number of positive samples

    """
    # load kitsune model
    print("evaluting", path)
    print("kitsune model path ", model_path)
    with open(model_path, "rb") as m:
        kitsune = pickle.load(m)

    # load pcap file
    counter = 0
    input_file = open(path, "r")
    input_file.readline()
    rmse_array = []
    d_rmse = []
    d_rel_rmse = []

    tbar = tqdm()

    feature_vector = input_file.readline()
    while feature_vector is not '':
        fv = feature_vector.rstrip().split(",")

        #remove label if there is one

        fv = fv[:100]

        fv = np.array(fv, dtype="float")

        squeezed_fv = squeeze_features(fv, precision)

        rmse = kitsune.process(fv)
        squeezed_rmse = kitsune.process(squeezed_fv)
        d_rmse.append(np.abs(rmse-squeezed_rmse))
        d_rel_rmse.append(np.abs(rmse-squeezed_rmse)/rmse)

        if rmse == 0:
            rmse_array.append(1e-2)
        else:
            rmse_array.append(rmse)
        counter += 1
        tbar.update(1)

        feature_vector = input_file.readline()
    tbar.close()
    if not threshold:
        # calculate kitsune threshold and detector threshold and number of positive_samples
        kitsune_threshold = calc_threshold(rmse_array)
        pos_kit = (rmse_array > kitsune_threshold).sum()

        d_rmse_threshold = calc_threshold(d_rmse)
        pos_d_rmse = (d_rmse > d_rmse_threshold).sum()

        d_rel_rmse_threshold = calc_threshold(d_rel_rmse)
        pos_d_rel_rmse = (d_rel_rmse > d_rel_rmse_threshold).sum()
        return kitsune_threshold, pos_kit, d_rmse_threshold, pos_d_rmse, d_rel_rmse_threshold, pos_d_rel_rmse
    else:
        pos_kit = (rmse_array > threshold[0]).sum()

        pos_d_rmse = (d_rmse > threshold[1]).sum()

        pos_d_rel_rmse = (d_rel_rmse > threshold[2]).sum()

        return pos_kit, pos_d_rmse, pos_d_rel_rmse


def calc_threshold(array):
    benignSample = np.log(array)
    mean = np.mean(benignSample)
    std = np.std(benignSample)
    threshold_std = np.exp(mean + 3 * std)
    threshold_max = max(array)
    threshold = min(threshold_max, threshold_std)
    return threshold



import csv
import numpy as np
from sklearn import metrics

def evaluation_metrics(scores_path, threshold_path):
    # Calulate the F1 score for Model using ground labels
    # Actual outputs from kitsune_score.csv
    y_output = csv.reader(open(scores_path, 'r'))
    y_output = list(y_output)
    y_output = [float(i[0]) for i in y_output]

    # let ground labels be 0(benign) for all pkts
    y_true = np.ones(len(y_output))

    # Threshold for F1 score
    threshold = csv.reader(open(threshold_path, 'r'))
    threshold = list(threshold)
    threshold = float(threshold[0][0])

    # Predicted labels
    y_pred = np.array([0 if i > threshold else 1 for i in y_output])

    # 0 -> benign and 1 -> malicious

    tn, fp, fn, tp = metrics.confusion_matrix(y_true, y_pred).ravel()
    precision = tp/(tp+fp)
    recall = tp/(tp+fn)
    f1 = 2*(precision*recall)/(precision+recall)

    print("True Negatives: ", tn)
    print("False Positives: ", fp)
    print("False Negatives: ", fn)
    print("True Positives: ", tp)

    print("Precision: ", precision)
    print("Recall: ", recall)
    print("F1 score: ", f1)

    return f1



# from kitsune_ids import parse_with_kitsune as p
# create a class AwesomeIDS that would have the parse fuction that basically the calls the parse_kitsune function from the parse_with_kitsune.py file 
# and then import the AwesomeIDS class in the __init__.py file and then call the parse function from the AwesomeIDS class in the __init__.py file

# from ..kitsune_ids import parse_with_kitsune

from ..kitsune_ids import parse_with_kitsune
from ..kitsune_ids.kitsune import *

# class AwesomeIDS:
#     """
#     This is an awesome IDS
#     """
#     def parse (self,pcap_file, output_file_name, add_label=False, write_prob=1, count=float('Inf'), parse_type="scapy", netstat_path=None, save_netstat=None, add_proto=False, add_time=False, netstat_log_file=None):
#         parse_with_kitsune.parse_kitsune(pcap_file, output_file_name, add_label, write_prob, count, parse_type, netstat_path, save_netstat, add_proto, add_time, netstat_log_file)

class AwesomeIDS:
    """
    This is an awesome IDS
    """
    # This is for parsing the data and extracting the features from the data .
    def parse (self,pcap_file, output_file_name, add_label=False, write_prob=1, count=float('Inf'), parse_type="scapy", netstat_path=None, save_netstat=None, add_proto=False, add_time=False, netstat_log_file=None):
        parse_with_kitsune.parse_kitsune(pcap_file, output_file_name, add_label, write_prob, count, parse_type, netstat_path, save_netstat, add_proto, add_time, netstat_log_file)

    # make a function that would train the model and then save the model in a file
    def train_model(self, params):
        train_normal(params) 

    def test_model(self, path, model_path, threshold=None, ignore_index=-1, out_image=None, meta_file=None, record_scores=False, y_true=None, record_prediction=False, load_prediction=False, plot_with_time=False):
        return eval_kitsune(path, model_path, threshold, ignore_index, out_image, meta_file, record_scores, y_true, record_prediction, load_prediction, plot_with_time)
